"""Unit test for Address class."""

import pytest

from xknx.exceptions import ConversionError
from xknx.telegram import AddressFilter
from xknx.telegram.address import GroupAddress, InternalGroupAddress


class TestAddressFilter:
    """Test class for Address."""

    def test_range_initialization(self) -> None:
        """Test Initialization of AddresFilter.Range."""
        assert AddressFilter.Range("*").get_range() == (0, 65535)
        assert AddressFilter.Range("5").get_range() == (5, 5)
        assert AddressFilter.Range("0").get_range() == (0, 0)
        assert AddressFilter.Range("3-5").get_range() == (3, 5)
        assert AddressFilter.Range("5-3").get_range() == (3, 5)
        assert AddressFilter.Range("-5").get_range() == (0, 5)
        assert AddressFilter.Range("5-").get_range() == (5, 65535)
        assert AddressFilter.Range("70-100").get_range() == (70, 100)

    def test_range_test(self) -> None:
        """Test matching within AddressFilter.Range."""
        range_filter = AddressFilter.Range("2-16")
        assert range_filter.match(10)
        assert range_filter.match(2)
        assert range_filter.match(16)
        assert not range_filter.match(1)
        assert not range_filter.match(17)

    def test_level_filter_test(self) -> None:
        """Test matching within AddressFilter.LevelFilter."""
        level_filter = AddressFilter.LevelFilter("2,4,8-10,13")
        assert not level_filter.match(1)
        assert level_filter.match(2)
        assert not level_filter.match(3)
        assert level_filter.match(4)
        assert not level_filter.match(5)
        assert level_filter.match(9)

    def test_address_filter_level3_3(self) -> None:
        """Test AddressFilter 3rd part of level3 addresses."""
        af1 = AddressFilter("1/2/3")
        assert af1.match("1/2/3")
        assert not af1.match("1/2/4")
        assert not af1.match("1/2/1")
        af2 = AddressFilter("1/2/2-3,5-")
        assert not af2.match("1/2/1")
        assert af2.match("1/2/3")
        assert not af2.match("1/2/4")
        assert af2.match("1/2/6")
        af3 = AddressFilter("1/2/*")
        assert af3.match("1/2/3")
        assert af3.match("1/2/5")

    def test_address_filter_level3_2(self) -> None:
        """Test AddressFilter 2nd part of level3 addresses."""
        af1 = AddressFilter("1/2/3")
        assert af1.match("1/2/3")
        assert not af1.match("1/3/3")
        assert not af1.match("1/1/3")
        af2 = AddressFilter("1/2-/3")
        assert not af2.match("1/1/3")
        assert af2.match("1/2/3")
        assert af2.match("1/5/3")
        af3 = AddressFilter("1/*/3")
        assert af3.match("1/4/3")
        assert af3.match("1/7/3")

    def test_address_filter_level3_1(self) -> None:
        """Test AddressFilter 1st part of level3 addresses."""
        af1 = AddressFilter("4/2/3")
        assert af1.match("4/2/3")
        assert not af1.match("2/2/3")
        assert not af1.match("10/2/3")
        af2 = AddressFilter("2-/4/3")
        assert not af2.match("1/4/3")
        assert af2.match("2/4/3")
        assert af2.match("10/4/3")
        af3 = AddressFilter("*/5/5")
        assert af3.match("2/5/5")
        assert af3.match("8/5/5")

    def test_address_filter_level2_2(self) -> None:
        """Test AddressFilter 2nd part of level2 addresses."""
        af1 = AddressFilter("2/3")
        assert af1.match("2/3")
        assert not af1.match("2/4")
        assert not af1.match("2/1")
        af2 = AddressFilter("2/3-4,7-")
        assert not af2.match("2/2")
        assert af2.match("2/3")
        assert not af2.match("2/6")
        assert af2.match("2/8")
        af3 = AddressFilter("2/*")
        assert af3.match("2/3")
        assert af3.match("2/5")

    def test_address_filter_level2_1(self) -> None:
        """Test AddressFilter 1st part of level2 addresses."""
        af1 = AddressFilter("4/2")
        assert af1.match("4/2")
        assert not af1.match("2/2")
        assert not af1.match("10/2")
        af2 = AddressFilter("2-3,8-/4")
        assert not af2.match("1/4")
        assert af2.match("2/4")
        assert not af2.match("7/4")
        assert af2.match("10/4")
        af3 = AddressFilter("*/5")
        assert af3.match("2/5")
        assert af3.match("8/5")

    def test_address_filter_free(self) -> None:
        """Test AddressFilter free format addresses."""
        af1 = AddressFilter("4")
        assert af1.match("4")
        assert not af1.match("1")
        assert not af1.match("10")
        af2 = AddressFilter("1,4,7-")
        assert af2.match("1")
        assert not af2.match("2")
        assert af2.match("4")
        assert not af2.match("6")
        assert af2.match("60")
        af3 = AddressFilter("*")
        assert af3.match("2")
        assert af3.match("8")

    def test_address_combined(self) -> None:
        """Test AddressFilter with complex filters."""
        af1 = AddressFilter("2-/2,3,5-/*")
        assert af1.match("2/3/8")
        assert af1.match("4/7/10")
        assert af1.match("2/7/10")
        assert not af1.match("1/7/10")
        assert not af1.match("2/4/10")
        assert not af1.match("2/1/10")

    def test_initialize_wrong_format(self) -> None:
        """Test if wrong address format raises exception."""
        with pytest.raises(ConversionError):
            AddressFilter("2-/2,3/4/5/1,5-/*")

    def test_adjust_range(self) -> None:
        """Test helper function _adjust_range."""
        assert (
            AddressFilter.Range._adjust_range(GroupAddress.MAX_FREE + 1)
            == GroupAddress.MAX_FREE
        )
        assert AddressFilter.Range._adjust_range(-1) == 0

    def test_internal_address_filter_exact(self) -> None:
        """Test AddressFilter for InternalGroupAddress."""
        af1 = AddressFilter("i-1")
        assert af1.match("i1")
        assert af1.match("i 1")
        assert af1.match("i-1")
        assert af1.match(InternalGroupAddress("i-1"))
        assert not af1.match("1")
        assert not af1.match(GroupAddress(1))

    def test_internal_address_filter_wildcard(self) -> None:
        """Test AddressFilter with wildcards for InternalGroupAddress."""
        af1 = AddressFilter("i-?")
        assert af1.match("i1")
        assert af1.match("i 2")
        assert af1.match("i-3")
        assert af1.match(InternalGroupAddress("i-4"))
        assert not af1.match("1")
        assert not af1.match(GroupAddress(1))
        assert not af1.match("i11")
        assert not af1.match("i 11")
        assert not af1.match("i-11")
        assert not af1.match(InternalGroupAddress("i-11"))

        af2 = AddressFilter("i-t?st")
        assert af2.match("it1st")
        assert af2.match("i t2st")
        assert af2.match("i-test")
        assert af2.match(InternalGroupAddress("i-test"))
        assert not af2.match("1")
        assert not af2.match(GroupAddress(1))
        assert not af2.match("i11")
        assert not af2.match("i tst")
        assert not af2.match("i-teest")
        assert not af2.match(InternalGroupAddress("i-tst"))

        af3 = AddressFilter("i-*")
        assert af3.match("i1")
        assert af3.match("i asdf")
        assert af3.match("i-3sdf")
        assert af3.match(InternalGroupAddress("i-4"))
        assert not af3.match("1")
        assert not af3.match(GroupAddress(1))
        assert af3.match("i11")
        assert af3.match("i 11??")
        assert af3.match("i-11*")
        assert af3.match(InternalGroupAddress("i-11"))

        af4 = AddressFilter("i-t*t")
        assert af4.match("it1t")
        assert af4.match("i t22t")
        assert af4.match("i-testt")
        assert af4.match("i-tt")
        assert af4.match(InternalGroupAddress("i-t333t"))
        assert not af4.match("1")
        assert not af4.match(GroupAddress(1))
        assert not af4.match("i testx")
        assert not af4.match("i-11test")
        assert not af4.match(InternalGroupAddress("i-11"))



"""Unit test for KNX/IP SessionAuthenticate objects."""

from xknx.knxip import KNXIPFrame, SessionAuthenticate


class TestKNXIPSessionAuthenticate:
    """Test class for KNX/IP SessionAuthenticate objects."""

    def test_session_authenticate(self) -> None:
        """Test parsing and streaming session authenticate KNX/IP packet."""
        message_authentication_code = bytes.fromhex(
            "1f 1d 59 ea 9f 12 a1 52"  # Message Authentication Code
            "e5 d9 72 7f 08 46 2c de"
        )
        raw = (
            bytes.fromhex(
                "06 10 09 53 00 18"  # KNXnet/IP header
                "00 01"  # User ID
            )
            + message_authentication_code
        )
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, SessionAuthenticate)
        assert knxipframe.body.user_id == 1
        assert (
            knxipframe.body.message_authentication_code == message_authentication_code
        )

        assert knxipframe.to_knx() == raw

        session_authenticate = SessionAuthenticate(
            user_id=1,
            message_authentication_code=message_authentication_code,
        )
        knxipframe2 = KNXIPFrame.init_from_body(session_authenticate)

        assert knxipframe2.to_knx() == raw



"""Example for switching a light on and off."""

import asyncio

from xknx import XKNX
from xknx.devices import Scene


async def main() -> None:
    """Connect to KNX/IP bus and run scene."""
    async with XKNX() as xknx:
        scene = Scene(xknx, name="Romantic", group_address="7/0/9", scene_number=23)
        xknx.devices.async_add(scene)

        await scene.run()


asyncio.run(main())



"""Module for KNX Telegrams."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum

from xknx.dpt import DPTBase, DPTComplexData, DPTEnumData

from .address import GroupAddress, IndividualAddress, InternalGroupAddress
from .apci import APCI
from .tpci import TPCI, TDataBroadcast, TDataGroup, TDataIndividual


class TelegramDirection(Enum):
    """Enum class for the communication direction of a telegram (from KNX bus or to KNX bus)."""

    INCOMING = "Incoming"
    OUTGOING = "Outgoing"


@dataclass(slots=True)
class TelegramDecodedData:
    """Context for a telegram."""

    transcoder: type[DPTBase]
    value: bool | int | float | str | DPTComplexData | DPTEnumData

    def __str__(self) -> str:
        """Return object as readable string."""
        return (
            f"{self.value}{' ' + self.transcoder.unit if self.transcoder.unit is not None else ''}"
            f" ({self.transcoder.dpt_name()})"
        )


@dataclass(slots=True)
class Telegram:
    """
    Data transfer object for KNX telegrams.

    Represents a message exchanged on the KNX bus between the business logic
    (Devices, Management, etc.) and the underlying KNX/IP abstraction layer.

    Attributes:
        destination_address: Target GroupAddress, IndividualAddress, or
            InternalGroupAddress.
        direction: Communication direction (INCOMING or OUTGOING).
        payload: APCi payload containing the actual data (e.g., GroupValueWrite,
            GroupValueResponse). None for control information only telegrams.
        source_address: IndividualAddress of the sender. When default of 0.0.0 is
            used, it will be set automatically when sent.
        tpci: Transport Layer Control Information (TDataBroadcast, TDataGroup, or
            TDataIndividual). If not provided, it will be automatically inferred
            based on destination_address type.
        decoded_data: Optional decoded version of the payload including the
            transcoder class and decoded value. Set externally by GroupAddressDPT
            for convenience when the payload has already been decoded.
        data_secure: Flag indicating if the telegram was sent or received as
            DataSecure. Set externally by CEMIHandler. None if not yet processed.

    """

    destination_address: GroupAddress | IndividualAddress | InternalGroupAddress
    direction: TelegramDirection = TelegramDirection.OUTGOING
    payload: APCI | None = None
    source_address: IndividualAddress = field(
        default_factory=lambda: IndividualAddress(0)
    )
    tpci: TPCI = None  # type: ignore[assignment]  # set by initializer or in __post_init__
    # set by GroupAddressDPT
    decoded_data: TelegramDecodedData | None = field(
        init=False, default=None, compare=False, hash=False
    )
    # flag if telegram was sent or received as DataSecure, set by CEMIHandler
    data_secure: bool | None = field(
        init=False, default=None, compare=False, hash=False
    )

    def __post_init__(self) -> None:
        """Initialize Telegram class."""
        if self.tpci is None:
            if isinstance(self.destination_address, GroupAddress):  # type: ignore[unreachable]
                if self.destination_address.raw == 0:
                    self.tpci = TDataBroadcast()
                else:
                    self.tpci = TDataGroup()
            elif isinstance(self.destination_address, IndividualAddress):
                self.tpci = TDataIndividual()
            else:  # InternalGroupAddress
                self.tpci = TDataGroup()

    def __str__(self) -> str:
        """Return object as readable string."""
        data = f'payload="{self.payload}"' if self.payload else f'tpci="{self.tpci}"'
        decoded_data = (
            f' data="{self.decoded_data}"' if self.decoded_data is not None else ""
        )
        return (
            "<Telegram "
            f'direction="{self.direction.value}" '
            f'source_address="{self.source_address}" '
            f'destination_address="{self.destination_address}" '
            f"{data}{decoded_data} />"
        )



"""Unit test for KNX/IP transport base class."""

from unittest.mock import Mock, patch

from xknx.io.transport import KNXIPTransport
from xknx.knxip import (
    HPAI,
    ConnectionStateRequest,
    ConnectionStateResponse,
    KNXIPFrame,
    KNXIPServiceType,
)


class TestKNXIPTransport:
    """Test class for KNXIPTransport base class."""

    @patch.multiple(KNXIPTransport, __abstractmethods__=set())
    def test_callback(self) -> None:
        """Test if callback is called correctly."""
        transport = KNXIPTransport()
        transport.callbacks = []
        callback_mock = Mock()
        # Registering callback
        callback_instance = transport.register_callback(
            callback_mock, [KNXIPServiceType.CONNECTIONSTATE_RESPONSE]
        )
        assert len(transport.callbacks) == 1
        # Handle KNX/IP frame with different service type
        wrong_service_type_frame = KNXIPFrame.init_from_body(
            ConnectionStateRequest(),
        )
        transport.handle_knxipframe(wrong_service_type_frame, HPAI())
        callback_mock.assert_not_called()
        # Handle KNX/IP frame with correct service type
        correct_service_type_frame = KNXIPFrame.init_from_body(
            ConnectionStateResponse(),
        )
        transport.handle_knxipframe(correct_service_type_frame, HPAI())
        callback_mock.assert_called_once_with(
            correct_service_type_frame, HPAI(), transport
        )
        # Unregistering callback
        callback_mock.reset_mock()
        transport.unregister_callback(callback_instance)
        assert not transport.callbacks
        transport.handle_knxipframe(correct_service_type_frame, HPAI())
        callback_mock.assert_not_called()
        # Registering callback for any service type
        callback_instance = transport.register_callback(callback_mock, None)
        transport.handle_knxipframe(correct_service_type_frame, HPAI())
        callback_mock.assert_called_once_with(
            correct_service_type_frame, HPAI(), transport
        )



"""Unit test for Light objects."""

from unittest.mock import Mock, patch

from xknx import XKNX
from xknx.devices import Light
from xknx.devices.light import ColorTemperatureType
from xknx.dpt import DPTArray, DPTBinary
from xknx.dpt.dpt_242 import XYYColor
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueRead, GroupValueWrite

from ..conftest import EventLoopClockAdvancer


class TestLight:
    """Class for testing Light objects."""

    #
    # TEST SUPPORT DIMMING
    #
    def test_supports_dimm_true(self) -> None:
        """Test supports_dimm attribute with a light with dimmer."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch="1/6/4",
            group_address_brightness="1/6/6",
        )
        assert light.supports_brightness

    def test_supports_dimm_false(self) -> None:
        """Test supports_dimm attribute with a Light without dimmer."""
        xknx = XKNX()
        light = Light(xknx, "Diningroom.Light_1", group_address_switch="1/6/4")
        assert not light.supports_brightness

    #
    # TEST SUPPORT COLOR
    #
    def test_supports_color_true(self) -> None:
        """Test supports_color true."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch="1/6/4",
            group_address_color="1/6/5",
        )
        assert light.supports_color

    def test_supports_color_false(self) -> None:
        """Test supports_color false."""
        xknx = XKNX()
        light = Light(xknx, "Diningroom.Light_1", group_address_switch="1/6/4")
        assert not light.supports_color

    def test_supports_individual_color_true(self) -> None:
        """Test supports_color true."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch_red="1/1/1",
            group_address_switch_red_state="1/1/2",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_switch_green="1/1/5",
            group_address_switch_green_state="1/1/6",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_switch_blue="1/1/9",
            group_address_switch_blue_state="1/1/10",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
        )
        assert light.supports_color

    def test_supports_individual_color_only_brightness_true(self) -> None:
        """Test supports_color true."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Individual colors only brightness",
            group_address_brightness_red="1/1/3",
            group_address_brightness_green="1/1/7",
            group_address_brightness_blue="1/1/11",
        )
        assert light.supports_color

    def test_supports_individual_color_false(self) -> None:
        """Test supports_color false."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch_red="1/1/1",
            group_address_switch_red_state="1/1/2",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_switch_green="1/1/5",
            group_address_switch_green_state="1/1/6",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
        )
        assert not light.supports_color

    #
    # TEST SUPPORT COLOR RGBW
    #
    def test_supports_rgbw_true(self) -> None:
        """Test supports_rgbw true."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch="1/6/4",
            group_address_rgbw="1/6/5",
            group_address_color="1/6/6",
        )
        assert light.supports_rgbw

    def test_supports_rgbw_false(self) -> None:
        """Test supports_color false."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch="1/6/4",
            group_address_color="1/6/6",
        )
        assert not light.supports_rgbw

    def test_supports_individual_rgbw_true(self) -> None:
        """Test supports_rgbw true."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch_red="1/1/1",
            group_address_switch_red_state="1/1/2",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_switch_green="1/1/5",
            group_address_switch_green_state="1/1/6",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_switch_blue="1/1/9",
            group_address_switch_blue_state="1/1/10",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
            group_address_switch_white="1/1/13",
            group_address_switch_white_state="1/1/14",
            group_address_brightness_white="1/1/15",
            group_address_brightness_white_state="1/1/16",
        )
        assert light.supports_rgbw

    def test_supports_individual_color_only_brightness_rgbw_true(self) -> None:
        """Test supports_color true."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Individual colors only brightness",
            group_address_brightness_red="1/1/3",
            group_address_brightness_green="1/1/7",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_white="1/1/12",
        )
        assert light.supports_rgbw

    def test_supports_individual_rgbw_false(self) -> None:
        """Test supports_color false."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch_red="1/1/1",
            group_address_switch_red_state="1/1/2",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_switch_green="1/1/5",
            group_address_switch_green_state="1/1/6",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_switch_blue="1/1/9",
            group_address_switch_blue_state="1/1/10",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
        )
        assert not light.supports_rgbw

    def test_supports_hs_color_true(self) -> None:
        """Test supports_hs_color true."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Hue and saturation",
            group_address_switch="1/6/4",
            group_address_hue="1/6/5",
            group_address_saturation="1/6/6",
        )
        assert light.supports_hs_color

    def test_supports_hs_color_false(self) -> None:
        """Test supports_hs_color false."""
        xknx = XKNX()
        light_hue = Light(
            xknx,
            "Light hue only",
            group_address_switch="1/6/4",
            group_address_hue="1/6/5",
        )
        assert not light_hue.supports_hs_color

        light_saturation = Light(
            xknx,
            "Light saturation only",
            group_address_switch="1/6/4",
            group_address_saturation="1/6/5",
        )
        assert not light_saturation.supports_hs_color

    def test_supports_xyy_color_true(self) -> None:
        """Test supports_xyy_color true."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch="1/6/4",
            group_address_xyy_color="1/6/5",
        )
        assert light.supports_xyy_color

    def test_supports_xyy_color_false(self) -> None:
        """Test supports_xyy_color false."""
        xknx = XKNX()
        light = Light(xknx, "Diningroom.Light_1", group_address_switch="1/6/4")
        assert not light.supports_xyy_color

    #
    # TEST SUPPORT TUNABLE WHITE
    #
    def test_supports_tw_yes(self) -> None:
        """Test supports_tw attribute with a light with tunable white function."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch="1/6/4",
            group_address_tunable_white="1/6/6",
        )
        assert light.supports_tunable_white

    def test_supports_tw_no(self) -> None:
        """Test supports_tw attribute with a Light without tunable white function."""
        xknx = XKNX()
        light = Light(xknx, "Diningroom.Light_1", group_address_switch="1/6/4")
        assert not light.supports_tunable_white

    #
    # TEST SUPPORT COLOR TEMPERATURE
    #
    def test_supports_color_temp_true(self) -> None:
        """Test supports_color_temp attribute with a light with color temperature function."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch="1/6/4",
            group_address_color_temperature="1/6/6",
        )
        assert light.supports_color_temperature

    def test_supports_color_temp_false(self) -> None:
        """Test supports_color_temp attribute with a Light without color temperature function."""
        xknx = XKNX()
        light = Light(xknx, "Diningroom.Light_1", group_address_switch="1/6/4")
        assert not light.supports_color_temperature

    #
    # SYNC
    #
    async def test_sync(self) -> None:
        """Test sync function / sending group reads to KNX bus."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/3/5",
            group_address_switch_state="1/2/3",
            group_address_brightness_state="1/2/5",
            group_address_color_state="1/2/6",
            group_address_xyy_color_state="1/2/4",
            group_address_tunable_white_state="1/2/7",
            group_address_color_temperature_state="1/2/8",
            group_address_rgbw_state="1/2/9",
        )
        expected_telegrams = 7

        await light.sync()
        assert xknx.telegrams.qsize() == expected_telegrams

        telegrams = [xknx.telegrams.get_nowait() for _ in range(expected_telegrams)]
        test_telegrams = [
            Telegram(
                destination_address=GroupAddress("1/2/3"), payload=GroupValueRead()
            ),
            Telegram(
                destination_address=GroupAddress("1/2/5"), payload=GroupValueRead()
            ),
            Telegram(
                destination_address=GroupAddress("1/2/6"), payload=GroupValueRead()
            ),
            Telegram(
                destination_address=GroupAddress("1/2/4"), payload=GroupValueRead()
            ),
            Telegram(
                destination_address=GroupAddress("1/2/9"), payload=GroupValueRead()
            ),
            Telegram(
                destination_address=GroupAddress("1/2/7"), payload=GroupValueRead()
            ),
            Telegram(
                destination_address=GroupAddress("1/2/8"), payload=GroupValueRead()
            ),
        ]
        for test_telegram in test_telegrams:
            assert test_telegram in telegrams

    async def test_sync_individual_color(self) -> None:
        """Test sync function / sending group reads to KNX bus. Testing with a Light without dimm functionality."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch_red="1/1/1",
            group_address_switch_red_state="1/1/2",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_switch_green="1/1/5",
            group_address_switch_green_state="1/1/6",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_switch_blue="1/1/9",
            group_address_switch_blue_state="1/1/10",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
            group_address_switch_white="1/1/13",
            group_address_switch_white_state="1/1/14",
            group_address_brightness_white="1/1/15",
            group_address_brightness_white_state="1/1/16",
        )
        await light.sync()

        assert xknx.telegrams.qsize() == 8

        telegrams = [xknx.telegrams.get_nowait() for _ in range(8)]

        test_telegrams = [
            Telegram(
                destination_address=GroupAddress("1/1/2"),
                payload=GroupValueRead(),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/4"),
                payload=GroupValueRead(),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/6"),
                payload=GroupValueRead(),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/8"),
                payload=GroupValueRead(),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/10"),
                payload=GroupValueRead(),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/12"),
                payload=GroupValueRead(),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/14"),
                payload=GroupValueRead(),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/16"),
                payload=GroupValueRead(),
            ),
        ]
        for test_telegram in test_telegrams:
            assert test_telegram in telegrams

    #
    # TEST SET ON
    #
    async def test_set_on(self) -> None:
        """Test switching on a Light."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_brightness="1/2/5",
        )
        await light.set_on()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )

    async def test_set_on_individual_color(self) -> None:
        """Test switching on a Light."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch_red="1/1/1",
            group_address_switch_red_state="1/1/2",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_switch_green="1/1/5",
            group_address_switch_green_state="1/1/6",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_switch_blue="1/1/9",
            group_address_switch_blue_state="1/1/10",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
            group_address_switch_white="1/1/13",
            group_address_switch_white_state="1/1/14",
            group_address_brightness_white="1/1/15",
            group_address_brightness_white_state="1/1/16",
        )
        assert light.state is None
        for color in light._iter_individual_colors():
            assert color.is_on is None

        await light.set_on()
        assert xknx.telegrams.qsize() == 4

        telegrams = [xknx.telegrams.get_nowait() for _ in range(4)]

        test_telegrams = [
            Telegram(
                destination_address=GroupAddress("1/1/1"),
                payload=GroupValueWrite(DPTBinary(True)),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/5"),
                payload=GroupValueWrite(DPTBinary(True)),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/9"),
                payload=GroupValueWrite(DPTBinary(True)),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/13"),
                payload=GroupValueWrite(DPTBinary(True)),
            ),
        ]
        for test_telegram in test_telegrams:
            assert test_telegram in telegrams

        for telegram in telegrams:
            light.process(telegram)

        assert light.state is True
        for color in light._iter_individual_colors():
            assert color.is_on is True

    async def test_set_on_individual_color_only_brightness(self) -> None:
        """Test switching on a Light."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Individual colors only brightness",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
            group_address_brightness_white="1/1/15",
            group_address_brightness_white_state="1/1/16",
        )
        assert light.state is None
        for color in light._iter_individual_colors():
            assert color.is_on is None

        await light.set_on()
        assert xknx.telegrams.qsize() == 4

        telegrams = [xknx.telegrams.get_nowait() for _ in range(4)]

        test_telegrams = [
            Telegram(
                destination_address=GroupAddress("1/1/3"),
                payload=GroupValueWrite(DPTArray((0xFF,))),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/7"),
                payload=GroupValueWrite(DPTArray((0xFF,))),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/11"),
                payload=GroupValueWrite(DPTArray((0xFF,))),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/15"),
                payload=GroupValueWrite(DPTArray((0xFF,))),
            ),
        ]
        for test_telegram in test_telegrams:
            assert test_telegram in telegrams

        for telegram in telegrams:
            light.process(telegram)

        assert light.state is True
        for color in light._iter_individual_colors():
            assert color.is_on is True

    #
    # TEST SET OFF
    #
    async def test_set_off(self) -> None:
        """Test switching off a Light."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_brightness="1/2/5",
        )
        await light.set_off()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(0)),
        )

        light.process(telegram)
        assert light.state is False

    async def test_set_off_individual_color(self) -> None:
        """Test switching off a Light."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch_red="1/1/1",
            group_address_switch_red_state="1/1/2",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_switch_green="1/1/5",
            group_address_switch_green_state="1/1/6",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_switch_blue="1/1/9",
            group_address_switch_blue_state="1/1/10",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
            group_address_switch_white="1/1/13",
            group_address_switch_white_state="1/1/14",
            group_address_brightness_white="1/1/15",
            group_address_brightness_white_state="1/1/16",
        )
        await light.set_off()
        assert xknx.telegrams.qsize() == 4

        telegrams = [xknx.telegrams.get_nowait() for _ in range(4)]

        test_telegrams = [
            Telegram(
                destination_address=GroupAddress("1/1/1"),
                payload=GroupValueWrite(DPTBinary(False)),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/5"),
                payload=GroupValueWrite(DPTBinary(False)),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/9"),
                payload=GroupValueWrite(DPTBinary(False)),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/13"),
                payload=GroupValueWrite(DPTBinary(False)),
            ),
        ]
        for test_telegram in test_telegrams:
            assert test_telegram in telegrams

        for telegram in telegrams:
            light.process(telegram)
        assert light.state is False
        for color in light._iter_individual_colors():
            assert color.is_on is False

    async def test_set_off_individual_color_only_brightness(self) -> None:
        """Test switching off a Light."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
            group_address_brightness_white="1/1/15",
            group_address_brightness_white_state="1/1/16",
        )
        await light.set_off()
        assert xknx.telegrams.qsize() == 4

        telegrams = [xknx.telegrams.get_nowait() for _ in range(4)]

        test_telegrams = [
            Telegram(
                destination_address=GroupAddress("1/1/3"),
                payload=GroupValueWrite(DPTArray((0,))),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/7"),
                payload=GroupValueWrite(DPTArray((0,))),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/11"),
                payload=GroupValueWrite(DPTArray((0,))),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/15"),
                payload=GroupValueWrite(DPTArray((0,))),
            ),
        ]
        for test_telegram in test_telegrams:
            assert test_telegram in telegrams

        for telegram in telegrams:
            light.process(telegram)
        assert light.state is False
        for color in light._iter_individual_colors():
            assert color.is_on is False

    #
    # TEST SET BRIGHTNESS
    #
    async def test_set_brightness(self) -> None:
        """Test setting the brightness of a Light."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_brightness="1/2/5",
        )
        await light.set_brightness(23)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTArray(23)),
        )

    async def test_set_brightness_not_dimmable(self) -> None:
        """Test setting the brightness of a non dimmable Light."""
        xknx = XKNX()
        light = Light(xknx, name="TestLight", group_address_switch="1/2/3")
        xknx.devices.async_add(light)

        with patch("logging.Logger.warning") as mock_warn:
            await light.set_brightness(23)
            assert xknx.telegrams.qsize() == 0
            mock_warn.assert_called_with(
                "Dimming not supported for device %s", "TestLight"
            )

    async def test_set_individual_color_with_gloabl_switch(self) -> None:
        """Test switching on and dimming a Light with global addresses."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch="1/0/0",
            group_address_brightness="1/0/1",
            group_address_switch_red="1/1/1",
            group_address_switch_red_state="1/1/2",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_switch_green="1/1/5",
            group_address_switch_green_state="1/1/6",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_switch_blue="1/1/9",
            group_address_switch_blue_state="1/1/10",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
            group_address_switch_white="1/1/13",
            group_address_switch_white_state="1/1/14",
            group_address_brightness_white="1/1/15",
            group_address_brightness_white_state="1/1/16",
        )
        assert light.state is None
        for color in light._iter_individual_colors():
            assert color.is_on is None
        # turn on
        await light.set_on()
        assert xknx.telegrams.qsize() == 1

        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/0/0"),
            payload=GroupValueWrite(DPTBinary(True)),
        )
        # brightness
        await light.set_brightness(23)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/0/1"),
            payload=GroupValueWrite(DPTArray(23)),
        )

    #
    # TEST SET COLOR
    #
    async def test_set_color(self) -> None:
        """Test setting the color of a Light."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_color="1/2/5",
        )
        xknx.devices.async_add(light)

        await light.set_color((23, 24, 25))
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTArray((23, 24, 25))),
        )
        xknx.devices.process(telegram)
        assert light.current_color == ((23, 24, 25), None)

    async def test_set_color_not_possible(self) -> None:
        """Test setting the color of a non light without color."""
        xknx = XKNX()
        light = Light(xknx, name="TestLight", group_address_switch="1/2/3")
        xknx.devices.async_add(light)

        with patch("logging.Logger.warning") as mock_warn:
            await light.set_color((23, 24, 25))
            assert xknx.telegrams.qsize() == 0
            mock_warn.assert_called_with(
                "Colors not supported for device %s", "TestLight"
            )

    async def test_set_individual_color(self) -> None:
        """Test setting the color of a Light."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch_red="1/1/1",
            group_address_switch_red_state="1/1/2",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_switch_green="1/1/5",
            group_address_switch_green_state="1/1/6",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_switch_blue="1/1/9",
            group_address_switch_blue_state="1/1/10",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
        )
        xknx.devices.async_add(light)

        await light.set_color([23, 24, 25])
        assert xknx.telegrams.qsize() == 3
        telegrams = [xknx.telegrams.get_nowait() for _ in range(3)]

        test_telegrams = [
            Telegram(
                destination_address=GroupAddress("1/1/3"),
                payload=GroupValueWrite(DPTArray(23)),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/7"),
                payload=GroupValueWrite(DPTArray(24)),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/11"),
                payload=GroupValueWrite(DPTArray(25)),
            ),
        ]
        for test_telegram in test_telegrams:
            assert test_telegram in telegrams

        xknx.devices.process(
            Telegram(
                destination_address=GroupAddress("1/1/4"),
                payload=GroupValueWrite(DPTArray(23)),
            )
        )
        xknx.devices.process(
            Telegram(
                destination_address=GroupAddress("1/1/8"),
                payload=GroupValueWrite(DPTArray(24)),
            )
        )
        xknx.devices.process(
            Telegram(
                destination_address=GroupAddress("1/1/12"),
                payload=GroupValueWrite(DPTArray(25)),
            )
        )
        assert light.current_color == ((23, 24, 25), None)

    async def test_set_individual_color_not_possible(self) -> None:
        """Test setting the color of a non light without color."""
        xknx = XKNX()
        light = Light(
            xknx,
            "TestLight",
            group_address_switch_red="1/1/1",
        )
        xknx.devices.async_add(light)

        with patch("logging.Logger.warning") as mock_warn:
            await light.set_color((23, 24, 25))
            assert xknx.telegrams.qsize() == 0
            mock_warn.assert_called_with(
                "Colors not supported for device %s", "TestLight"
            )

    #
    # TEST SET COLOR AS RGBW
    #
    async def test_set_color_rgbw(self) -> None:
        """Test setting RGBW value of a Light."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_color="1/2/4",
            group_address_rgbw="1/2/5",
        )
        xknx.devices.async_add(light)

        await light.set_color((23, 24, 25), 26)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTArray((23, 24, 25, 26, 0, 15))),
        )
        xknx.devices.process(telegram)
        assert light.current_color == ((23, 24, 25), 26)

    async def test_set_color_rgbw_not_possible(self) -> None:
        """Test setting RGBW value of a non light without color."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_color="1/2/4",
        )
        xknx.devices.async_add(light)

        with patch("logging.Logger.warning") as mock_warn:
            await light.set_color((23, 24, 25), 26)

            assert xknx.telegrams.qsize() == 0
            mock_warn.assert_called_with(
                "RGBW not supported for device %s", "TestLight"
            )

    async def test_set_individual_color_rgbw(self) -> None:
        """Test setting RGBW value of a Light."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch_red="1/1/1",
            group_address_switch_red_state="1/1/2",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_switch_green="1/1/5",
            group_address_switch_green_state="1/1/6",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_switch_blue="1/1/9",
            group_address_switch_blue_state="1/1/10",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
            group_address_switch_white="1/1/13",
            group_address_switch_white_state="1/1/14",
            group_address_brightness_white="1/1/15",
            group_address_brightness_white_state="1/1/16",
        )
        xknx.devices.async_add(light)

        await light.set_color([23, 24, 25], white=26)
        assert xknx.telegrams.qsize() == 4
        telegrams = [xknx.telegrams.get_nowait() for _ in range(4)]

        test_telegrams = [
            Telegram(
                destination_address=GroupAddress("1/1/3"),
                payload=GroupValueWrite(DPTArray(23)),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/7"),
                payload=GroupValueWrite(DPTArray(24)),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/11"),
                payload=GroupValueWrite(DPTArray(25)),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/15"),
                payload=GroupValueWrite(DPTArray(26)),
            ),
        ]
        for test_telegram in test_telegrams:
            assert test_telegram in telegrams

        xknx.devices.process(
            Telegram(
                destination_address=GroupAddress("1/1/4"),
                payload=GroupValueWrite(DPTArray(23)),
            )
        )
        xknx.devices.process(
            Telegram(
                destination_address=GroupAddress("1/1/8"),
                payload=GroupValueWrite(DPTArray(24)),
            )
        )
        xknx.devices.process(
            Telegram(
                destination_address=GroupAddress("1/1/12"),
                payload=GroupValueWrite(DPTArray(25)),
            )
        )
        xknx.devices.process(
            Telegram(
                destination_address=GroupAddress("1/1/16"),
                payload=GroupValueWrite(DPTArray(26)),
            )
        )
        assert light.current_color == ((23, 24, 25), 26)

    async def test_set_individual_color_rgbw_not_possible(self) -> None:
        """Test setting RGBW value of a non light without color."""
        xknx = XKNX()
        light = Light(
            xknx,
            "TestLight",
            group_address_switch_red="1/1/1",
            group_address_switch_red_state="1/1/2",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_switch_green="1/1/5",
            group_address_switch_green_state="1/1/6",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_switch_blue="1/1/9",
            group_address_switch_blue_state="1/1/10",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
        )
        xknx.devices.async_add(light)

        with patch("logging.Logger.warning") as mock_warn:
            await light.set_color((23, 24, 25), 26)

            assert xknx.telegrams.qsize() == 0
            mock_warn.assert_called_with(
                "RGBW not supported for device %s", "TestLight"
            )

    #
    # TEST SET COLOR AS HS
    #
    async def test_set_hs_color(self) -> None:
        """Test setting HS value of a Light."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_hue="1/2/4",
            group_address_saturation="1/2/5",
        )
        xknx.devices.async_add(light)

        await light.set_hs_color((359, 99))
        assert xknx.telegrams.qsize() == 2

        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/4"),
            payload=GroupValueWrite(DPTArray((0xFE,))),
        )
        xknx.devices.process(telegram)

        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTArray((0xFC,))),
        )
        xknx.devices.process(telegram)

        assert light.current_hs_color == (359, 99)

        # change only one
        await light.set_hs_color((18, 99))
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/4"),
            payload=GroupValueWrite(DPTArray((0x0D,))),
        )
        xknx.devices.process(telegram)

        await light.set_hs_color((18, 3))
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTArray((0x08,))),
        )
        xknx.devices.process(telegram)

        # call set_hs_color with current color shall trigger both values
        await light.set_hs_color((18, 3))
        assert xknx.telegrams.qsize() == 2

        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/4"),
            payload=GroupValueWrite(DPTArray((0x0D,))),
        )
        xknx.devices.process(telegram)

        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTArray((0x08,))),
        )
        xknx.devices.process(telegram)

    async def test_set_hs_color_not_possible(self) -> None:
        """Test setting HS value of a light not supporting it."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_color="1/2/4",
        )
        xknx.devices.async_add(light)

        with patch("logging.Logger.warning") as mock_warn:
            await light.set_hs_color((22, 25))

            assert xknx.telegrams.qsize() == 0
            mock_warn.assert_called_with(
                "HS-color not supported for device %s", "TestLight"
            )

    #
    # TEST SET COLOR AS XYY
    #
    async def test_set_xyy_color(self) -> None:
        """Test setting XYY value of a Light."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_xyy_color="1/2/4",
        )
        xknx.devices.async_add(light)

        await light.set_xyy_color(XYYColor((0.52, 0.31), 25))
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/4"),
            payload=GroupValueWrite(DPTArray((0x85, 0x1E, 0x4F, 0x5C, 0x19, 0x03))),
        )
        xknx.devices.process(telegram)
        assert light.current_xyy_color == XYYColor((0.52, 0.31), 25)

    async def test_set_xyy_color_not_possible(self) -> None:
        """Test setting XYY value of a light not supporting it."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_color="1/2/4",
        )
        xknx.devices.async_add(light)

        with patch("logging.Logger.warning") as mock_warn:
            await light.set_xyy_color(XYYColor((0.5, 0.3), 25))

            assert xknx.telegrams.qsize() == 0
            mock_warn.assert_called_with(
                "XYY-color not supported for device %s", "TestLight"
            )

    #
    # TEST SET TUNABLE WHITE
    #
    async def test_set_tw(self) -> None:
        """Test setting the tunable white value of a Light."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_tunable_white="1/2/5",
        )
        xknx.devices.async_add(light)

        await light.set_tunable_white(23)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTArray(23)),
        )

    async def test_set_tw_unsupported(self) -> None:
        """Test setting the tunable white value of a non tw Light."""
        xknx = XKNX()
        light = Light(xknx, name="TestLight", group_address_switch="1/2/3")
        xknx.devices.async_add(light)

        with patch("logging.Logger.warning") as mock_warn:
            await light.set_tunable_white(23)
            assert xknx.telegrams.qsize() == 0
            mock_warn.assert_called_with(
                "Tunable white not supported for device %s", "TestLight"
            )

    #
    # TEST SET COLOR TEMPERATURE
    #
    async def test_set_color_temp(self) -> None:
        """Test setting the color temperature value of a Light."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_color_temperature="1/2/5",
        )
        xknx.devices.async_add(light)

        await light.set_color_temperature(4000)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(
                DPTArray(
                    (
                        0x0F,
                        0xA0,
                    )
                )
            ),
        )

    async def test_set_color_temp_float(self) -> None:
        """Test setting the float color temperature value of a Light."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_color_temperature="1/2/5",
            color_temperature_type=ColorTemperatureType.FLOAT_2_BYTE,
        )
        xknx.devices.async_add(light)

        await light.set_color_temperature(4000)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(
                DPTArray(
                    (
                        0x46,
                        0x1A,
                    )
                )
            ),
        )

    async def test_set_color_temp_unsupported(self) -> None:
        """Test setting the color temperature value of an unsupported Light."""
        xknx = XKNX()
        light = Light(xknx, name="TestLight", group_address_switch="1/2/3")
        xknx.devices.async_add(light)

        with patch("logging.Logger.warning") as mock_warn:
            await light.set_color_temperature(4000)
            assert xknx.telegrams.qsize() == 0
            mock_warn.assert_called_with(
                "Absolute Color Temperature not supported for device %s", "TestLight"
            )

    #
    # TEST PROCESS
    #
    async def test_process_switch(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if switch position is processed correctly."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_brightness="1/2/5",
        )
        xknx.devices.async_add(light)

        assert light.state is None

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        light.process(telegram)
        assert light.state is True

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(0)),
        )
        light.process(telegram)
        assert light.state is False

    async def test_process_color_switch(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if switch position is processed correctly."""
        xknx = XKNX()
        light = Light(
            xknx,
            "TestLight",
            group_address_switch_red="1/1/1",
            group_address_switch_red_state="1/1/2",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_switch_green="1/1/5",
            group_address_switch_green_state="1/1/6",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_switch_blue="1/1/9",
            group_address_switch_blue_state="1/1/10",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
        )
        xknx.devices.async_add(light)

        assert light.state is None

        telegram = Telegram(
            destination_address=GroupAddress("1/1/2"),
            payload=GroupValueWrite(DPTBinary(True)),
        )
        light.process(telegram)
        assert light.state is True

        telegram = Telegram(
            destination_address=GroupAddress("1/1/2"),
            payload=GroupValueWrite(DPTBinary(False)),
        )
        light.process(telegram)
        assert light.state is False

    async def test_process_switch_callback(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if callback is called."""

        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_brightness="1/2/5",
        )
        xknx.devices.async_add(light)

        after_update_callback = Mock()
        light.register_device_updated_cb(after_update_callback)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        light.process(telegram)
        after_update_callback.assert_called_with(light)

    async def test_process_dimm(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if brightness is processed."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_brightness="1/2/5",
        )
        xknx.devices.async_add(light)

        assert light.current_brightness is None

        telegram = Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTArray(23)),
        )
        light.process(telegram)
        assert light.current_brightness == 23

    async def test_process_dimm_wrong_payload(self) -> None:
        """Test process wrong telegrams. (wrong payload type)."""
        xknx = XKNX()
        cb_mock = Mock()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_brightness="1/2/5",
            device_updated_cb=cb_mock,
        )
        xknx.devices.async_add(light)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        with patch("logging.Logger.warning") as log_mock:
            light.process(telegram)
            log_mock.assert_called_once()
            cb_mock.assert_not_called()

    async def test_process_dimm_payload_invalid_length(self) -> None:
        """Test process wrong telegrams. (wrong payload length)."""
        xknx = XKNX()
        cb_mock = Mock()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_brightness="1/2/5",
            device_updated_cb=cb_mock,
        )
        xknx.devices.async_add(light)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTArray((23, 24))),
        )
        with patch("logging.Logger.warning") as log_mock:
            light.process(telegram)
            log_mock.assert_called_once()
            cb_mock.assert_not_called()

    async def test_process_color(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if color is processed."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_color="1/2/5",
        )
        xknx.devices.async_add(light)

        assert light.current_color == (None, None)
        telegram = Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTArray((23, 24, 25))),
        )
        light.process(telegram)
        assert light.current_color == ((23, 24, 25), None)

    async def test_process_individual_color(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if color is processed."""
        xknx = XKNX()
        light = Light(
            xknx,
            "TestLight",
            group_address_switch_red="1/1/1",
            group_address_switch_red_state="1/1/2",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_switch_green="1/1/5",
            group_address_switch_green_state="1/1/6",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_switch_blue="1/1/9",
            group_address_switch_blue_state="1/1/10",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
        )
        xknx.devices.async_add(light)

        assert light.current_color == (None, None)

        telegrams = [
            Telegram(
                destination_address=GroupAddress("1/1/4"),
                payload=GroupValueWrite(DPTArray(42)),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/8"),
                payload=GroupValueWrite(DPTArray(43)),
            ),
            Telegram(
                destination_address=GroupAddress("1/1/12"),
                payload=GroupValueWrite(DPTArray(44)),
            ),
        ]

        for telegram in telegrams:
            light.process(telegram)
        assert light.current_color == ((42, 43, 44), None)

    async def test_process_color_rgbw(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if RGBW is processed."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_color="1/2/4",
            group_address_rgbw="1/2/5",
        )
        xknx.devices.async_add(light)

        assert light.current_color == (None, None)
        telegram = Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTArray((23, 24, 25, 26, 0, 15))),
        )
        light.process(telegram)
        assert light.current_color == ((23, 24, 25), 26)

    async def test_process_individual_color_rgbw(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if RGBW is processed."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Diningroom.Light_1",
            group_address_switch_red="1/1/1",
            group_address_switch_red_state="1/1/2",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_switch_green="1/1/5",
            group_address_switch_green_state="1/1/6",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_switch_blue="1/1/9",
            group_address_switch_blue_state="1/1/10",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
            group_address_switch_white="1/1/13",
            group_address_switch_white_state="1/1/14",
            group_address_brightness_white="1/1/15",
            group_address_brightness_white_state="1/1/16",
        )
        xknx.devices.async_add(light)

        assert light.current_color == (None, None)
        telegram = Telegram(
            destination_address=GroupAddress("1/1/16"),
            payload=GroupValueWrite(DPTArray(42)),
        )
        light.process(telegram)
        assert light.current_color == (None, 42)

    async def test_process_individual_color_debouncer(
        self, time_travel: EventLoopClockAdvancer
    ) -> None:
        """Test the debouncer for individual color lights."""
        xknx = XKNX()
        rgb_callback = Mock()
        rgbw_callback = Mock()

        rgb_light = Light(
            xknx,
            "TestRGBLight",
            group_address_switch_red="1/1/1",
            group_address_switch_red_state="1/1/2",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_switch_green="1/1/5",
            group_address_switch_green_state="1/1/6",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_switch_blue="1/1/9",
            group_address_switch_blue_state="1/1/10",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
            device_updated_cb=rgb_callback,
        )
        xknx.devices.async_add(rgb_light)
        rgbw_light = Light(
            xknx,
            "TestRGBWLight",
            group_address_switch="1/1/0",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
            group_address_brightness_white="1/1/15",
            group_address_brightness_white_state="1/1/16",
            device_updated_cb=rgbw_callback,
        )
        xknx.devices.async_add(rgbw_light)

        assert rgb_light.current_color == (None, None)
        assert rgbw_light.current_color == (None, None)

        red = Telegram(
            destination_address=GroupAddress("1/1/4"),
            payload=GroupValueWrite(DPTArray(42)),
        )
        green = Telegram(
            destination_address=GroupAddress("1/1/8"),
            payload=GroupValueWrite(DPTArray(43)),
        )
        blue = Telegram(
            destination_address=GroupAddress("1/1/12"),
            payload=GroupValueWrite(DPTArray(44)),
        )
        white = Telegram(
            destination_address=GroupAddress("1/1/16"),
            payload=GroupValueWrite(DPTArray(50)),
        )

        xknx.devices.process(red)
        rgb_callback.assert_not_called()
        rgbw_callback.assert_not_called()
        xknx.devices.process(green)
        rgb_callback.assert_not_called()
        rgbw_callback.assert_not_called()
        xknx.devices.process(blue)
        rgb_callback.assert_called_once()
        rgbw_callback.assert_not_called()
        xknx.devices.process(white)
        rgbw_callback.assert_called_once()

        assert rgb_light.current_color == ((42, 43, 44), None)
        assert rgbw_light.current_color == ((42, 43, 44), 50)

        rgb_callback.reset_mock()
        rgbw_callback.reset_mock()
        # second time with only 2 telegrams
        xknx.devices.process(red)
        rgb_callback.assert_not_called()
        rgbw_callback.assert_not_called()
        xknx.devices.process(green)
        rgb_callback.assert_not_called()
        rgbw_callback.assert_not_called()
        await time_travel(Light.DEBOUNCE_TIMEOUT / 2)
        rgb_callback.assert_not_called()
        rgbw_callback.assert_not_called()
        await time_travel(Light.DEBOUNCE_TIMEOUT / 2)
        rgb_callback.assert_called_once()
        rgbw_callback.assert_called_once()

        assert rgb_light.current_color == ((42, 43, 44), None)
        assert rgbw_light.current_color == ((42, 43, 44), 50)

    async def test_process_hs_color(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if color is processed."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_hue="1/2/4",
            group_address_hue_state="1/4/4",
            group_address_saturation="1/2/5",
            group_address_saturation_state="1/4/5",
        )
        assert light.current_hs_color is None
        # initialize hue
        light.process(
            Telegram(
                destination_address=GroupAddress("1/4/4"),
                payload=GroupValueWrite(DPTArray((0x2E,))),
            )
        )
        assert light.current_hs_color is None
        # initialize saturation
        light.process(
            Telegram(
                destination_address=GroupAddress("1/4/5"),
                payload=GroupValueWrite(DPTArray((0x55,))),
            )
        )
        assert light.current_hs_color == (65, 33)

    async def test_process_xyy_color(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if color is processed."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_xyy_color="1/2/5",
        )
        assert light.current_xyy_color is None
        # initial with invalid brightness
        light.process(
            Telegram(
                destination_address=GroupAddress("1/2/5"),
                payload=GroupValueWrite(DPTArray((0x2E, 0x14, 0x40, 0x00, 0x55, 0x02))),
            )
        )
        assert light.current_xyy_color == XYYColor((0.18, 0.25), None)
        # add valid brightness
        light.process(
            Telegram(
                destination_address=GroupAddress("1/2/5"),
                payload=GroupValueWrite(DPTArray((0x2E, 0x14, 0x40, 0x00, 0x55, 0x03))),
            )
        )
        assert light.current_xyy_color == XYYColor((0.18, 0.25), 85)
        # invalid color
        light.process(
            Telegram(
                destination_address=GroupAddress("1/2/5"),
                payload=GroupValueWrite(DPTArray((0xD1, 0xEB, 0xB0, 0xA3, 0xA5, 0x01))),
            )
        )
        assert light.current_xyy_color == XYYColor((0.18, 0.25), 165)
        # both valid
        light.process(
            Telegram(
                destination_address=GroupAddress("1/2/5"),
                payload=GroupValueWrite(DPTArray((0xD1, 0xEB, 0xB0, 0xA3, 0xA5, 0x03))),
            )
        )
        assert light.current_xyy_color == XYYColor((0.82, 0.69), 165)
        # invalid brightness
        light.process(
            Telegram(
                destination_address=GroupAddress("1/2/5"),
                payload=GroupValueWrite(DPTArray((0x2E, 0x14, 0x40, 0x00, 0x00, 0x02))),
            )
        )
        assert light.current_xyy_color == XYYColor((0.18, 0.25), 165)

    async def test_process_tunable_white(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if tunable white is processed."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_tunable_white="1/2/5",
        )
        assert light.current_tunable_white is None

        telegram = Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTArray(23)),
        )
        light.process(telegram)
        assert light.current_tunable_white == 23

    async def test_process_tunable_white_wrong_payload(self) -> None:
        """Test process wrong telegrams. (wrong payload type)."""
        xknx = XKNX()
        cb_mock = Mock()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_tunable_white="1/2/5",
            device_updated_cb=cb_mock,
        )
        telegram = Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        with patch("logging.Logger.warning") as log_mock:
            light.process(telegram)
            log_mock.assert_called_once()
            cb_mock.assert_not_called()

    async def test_process_tunable_white_payload_invalid_length(self) -> None:
        """Test process wrong telegrams. (wrong payload length)."""
        xknx = XKNX()
        cb_mock = Mock()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_tunable_white="1/2/5",
            device_updated_cb=cb_mock,
        )
        telegram = Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTArray((23, 24))),
        )
        with patch("logging.Logger.warning") as log_mock:
            light.process(telegram)
            log_mock.assert_called_once()
            cb_mock.assert_not_called()

    async def test_process_color_temperature(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if color temperature is processed."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_color_temperature="1/2/5",
        )
        assert light.current_color_temperature is None

        telegram = Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(
                DPTArray(
                    (
                        0x0F,
                        0xA0,
                    )
                )
            ),
        )
        light.process(telegram)
        assert light.current_color_temperature == 4000

    async def test_process_color_temperature_wrong_payload(self) -> None:
        """Test process wrong telegrams. (wrong payload type)."""
        xknx = XKNX()
        cb_mock = Mock()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_color_temperature="1/2/5",
            device_updated_cb=cb_mock,
        )
        telegram = Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        with patch("logging.Logger.warning") as log_mock:
            light.process(telegram)
            log_mock.assert_called_once()
            cb_mock.assert_not_called()

    async def test_process_color_temperature_payload_invalid_length(self) -> None:
        """Test process wrong telegrams. (wrong payload length)."""
        xknx = XKNX()
        cb_mock = Mock()
        light = Light(
            xknx,
            name="TestLight",
            group_address_switch="1/2/3",
            group_address_color_temperature="1/2/5",
            device_updated_cb=cb_mock,
        )
        telegram = Telegram(
            destination_address=GroupAddress("1/2/5"),
            payload=GroupValueWrite(DPTArray(23)),
        )
        with patch("logging.Logger.warning") as log_mock:
            light.process(telegram)
            log_mock.assert_called_once()
            cb_mock.assert_not_called()

    def test_has_group_address(self) -> None:
        """Test has_group_address."""
        xknx = XKNX()
        light = Light(
            xknx,
            "Office.Light_1",
            group_address_switch="1/7/1",
            group_address_switch_state="1/7/2",
            group_address_brightness="1/7/3",
            group_address_brightness_state="1/7/4",
            group_address_color="1/7/5",
            group_address_color_state="1/7/6",
            group_address_tunable_white="1/7/7",
            group_address_tunable_white_state="1/7/8",
            group_address_color_temperature="1/7/9",
            group_address_color_temperature_state="1/7/10",
            group_address_rgbw="1/7/11",
            group_address_rgbw_state="1/7/12",
            group_address_hue="1/7/81",
            group_address_hue_state="1/7/82",
            group_address_saturation="1/7/83",
            group_address_saturation_state="1/7/84",
            group_address_xyy_color="1/7/13",
            group_address_xyy_color_state="1/7/14",
            group_address_switch_red="1/1/1",
            group_address_switch_red_state="1/1/2",
            group_address_brightness_red="1/1/3",
            group_address_brightness_red_state="1/1/4",
            group_address_switch_green="1/1/5",
            group_address_switch_green_state="1/1/6",
            group_address_brightness_green="1/1/7",
            group_address_brightness_green_state="1/1/8",
            group_address_switch_blue="1/1/9",
            group_address_switch_blue_state="1/1/10",
            group_address_brightness_blue="1/1/11",
            group_address_brightness_blue_state="1/1/12",
            group_address_switch_white="1/1/13",
            group_address_switch_white_state="1/1/14",
            group_address_brightness_white="1/1/15",
            group_address_brightness_white_state="1/1/16",
        )
        assert light.has_group_address(GroupAddress("1/7/1"))
        assert light.has_group_address(GroupAddress("1/7/2"))
        assert light.has_group_address(GroupAddress("1/7/3"))
        assert light.has_group_address(GroupAddress("1/7/4"))
        assert light.has_group_address(GroupAddress("1/7/5"))
        assert light.has_group_address(GroupAddress("1/7/6"))
        assert light.has_group_address(GroupAddress("1/7/7"))
        assert light.has_group_address(GroupAddress("1/7/8"))
        assert light.has_group_address(GroupAddress("1/7/9"))
        assert light.has_group_address(GroupAddress("1/7/10"))
        assert light.has_group_address(GroupAddress("1/7/11"))
        assert light.has_group_address(GroupAddress("1/7/12"))
        # hue
        assert light.has_group_address(GroupAddress("1/7/81"))
        assert light.has_group_address(GroupAddress("1/7/82"))
        # saturation
        assert light.has_group_address(GroupAddress("1/7/83"))
        assert light.has_group_address(GroupAddress("1/7/84"))
        # xyy
        assert light.has_group_address(GroupAddress("1/7/13"))
        assert light.has_group_address(GroupAddress("1/7/14"))
        # individual
        assert light.has_group_address(GroupAddress("1/1/1"))
        assert light.has_group_address(GroupAddress("1/1/2"))
        assert light.has_group_address(GroupAddress("1/1/3"))
        assert light.has_group_address(GroupAddress("1/1/4"))
        assert light.has_group_address(GroupAddress("1/1/5"))
        assert light.has_group_address(GroupAddress("1/1/6"))
        assert light.has_group_address(GroupAddress("1/1/7"))
        assert light.has_group_address(GroupAddress("1/1/8"))
        assert light.has_group_address(GroupAddress("1/1/9"))
        assert light.has_group_address(GroupAddress("1/1/10"))
        assert light.has_group_address(GroupAddress("1/1/11"))
        assert light.has_group_address(GroupAddress("1/1/12"))
        assert light.has_group_address(GroupAddress("1/1/13"))
        assert light.has_group_address(GroupAddress("1/1/14"))
        assert light.has_group_address(GroupAddress("1/1/15"))
        assert light.has_group_address(GroupAddress("1/1/16"))

        assert not light.has_group_address(GroupAddress("1/7/15"))



"""Example for Fan device."""

import asyncio

from xknx import XKNX
from xknx.devices import Fan


async def main() -> None:
    """Connect to KNX/IP bus, control a fan, and turn it off afterwards."""
    xknx = XKNX()
    await xknx.start()

    fan = Fan(
        xknx,
        name="TestFan",
        group_address_switch="1/0/12",
        group_address_speed="1/0/14",
        max_step=3,
    )
    xknx.devices.async_add(fan)

    # Turn on the fan
    await fan.turn_on()

    # Set fan speed in steps
    for step in range(1, fan.max_step + 1):
        await fan.set_speed(step)
        await asyncio.sleep(1)

    # Turn off the fan
    await fan.turn_off()

    await xknx.stop()


asyncio.run(main())



"""Unit test for KNX/IP ConnectRequests."""

import pytest

from xknx.exceptions import CouldNotParseKNXIP
from xknx.knxip import (
    HPAI,
    ConnectRequest,
    ConnectRequestInformation,
    ConnectRequestType,
    KNXIPFrame,
    TunnellingLayer,
)
from xknx.telegram import IndividualAddress


class TestKNXIPConnectRequest:
    """Test class for KNX/IP ConnectRequests."""

    def test_tunnel_connect_request(self) -> None:
        """Test parsing and streaming connection request KNX/IP packet."""
        raw = bytes.fromhex(
            "06 10 02 05 00 1A 08 01 C0 A8 2A 01 84 95 08 01"
            "C0 A8 2A 01 CC A9 04 04 02 00"
        )
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, ConnectRequest)
        assert knxipframe.body.control_endpoint == HPAI(
            ip_addr="192.168.42.1", port=33941
        )
        assert knxipframe.body.data_endpoint == HPAI(ip_addr="192.168.42.1", port=52393)
        assert (
            knxipframe.body.cri.connection_type is ConnectRequestType.TUNNEL_CONNECTION
        )
        assert knxipframe.body.cri.knx_layer is TunnellingLayer.DATA_LINK_LAYER
        assert knxipframe.body.cri.individual_address is None

        cri = ConnectRequestInformation(
            connection_type=ConnectRequestType.TUNNEL_CONNECTION,
            knx_layer=TunnellingLayer.DATA_LINK_LAYER,
            individual_address=None,
        )
        connect_request = ConnectRequest(
            control_endpoint=HPAI(ip_addr="192.168.42.1", port=33941),
            data_endpoint=HPAI(ip_addr="192.168.42.1", port=52393),
            cri=cri,
        )
        knxipframe2 = KNXIPFrame.init_from_body(connect_request)

        assert knxipframe2.to_knx() == raw

    def test_mgmt_connect_request(self) -> None:
        """Test parsing and streaming connection request KNX/IP packet."""
        raw = bytes.fromhex(
            "06 10 02 05 00 18 08 01 C0 A8 2A 01 84 95 08 01 C0 A8 2A 01 CC A9 02 03"
        )
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, ConnectRequest)
        assert knxipframe.body.control_endpoint == HPAI(
            ip_addr="192.168.42.1", port=33941
        )
        assert knxipframe.body.data_endpoint == HPAI(ip_addr="192.168.42.1", port=52393)
        assert (
            knxipframe.body.cri.connection_type
            is ConnectRequestType.DEVICE_MGMT_CONNECTION
        )
        assert knxipframe.body.cri.individual_address is None

        cri = ConnectRequestInformation(
            connection_type=ConnectRequestType.DEVICE_MGMT_CONNECTION,
            individual_address=None,
        )
        connect_request = ConnectRequest(
            control_endpoint=HPAI(ip_addr="192.168.42.1", port=33941),
            data_endpoint=HPAI(ip_addr="192.168.42.1", port=52393),
            cri=cri,
        )
        knxipframe2 = KNXIPFrame.init_from_body(connect_request)

        assert knxipframe2.to_knx() == raw

    def test_connect_request_extended_cri(self) -> None:
        """Test parsing and streaming connection request KNX/IP packet."""
        raw = bytes.fromhex(
            "06 10 02 05 00 1C 08 01 C0 A8 2A 01 84 95 08 01"
            "C0 A8 2A 01 CC A9 06 04 02 00 01 02"
        )
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, ConnectRequest)
        assert knxipframe.body.control_endpoint == HPAI(
            ip_addr="192.168.42.1", port=33941
        )
        assert knxipframe.body.data_endpoint == HPAI(ip_addr="192.168.42.1", port=52393)
        assert (
            knxipframe.body.cri.connection_type is ConnectRequestType.TUNNEL_CONNECTION
        )
        assert knxipframe.body.cri.knx_layer is TunnellingLayer.DATA_LINK_LAYER
        assert knxipframe.body.cri.individual_address == IndividualAddress("0.1.2")

        cri = ConnectRequestInformation(
            connection_type=ConnectRequestType.TUNNEL_CONNECTION,
            knx_layer=TunnellingLayer.DATA_LINK_LAYER,
            individual_address=IndividualAddress("0.1.2"),
        )
        connect_request = ConnectRequest(
            control_endpoint=HPAI(ip_addr="192.168.42.1", port=33941),
            data_endpoint=HPAI(ip_addr="192.168.42.1", port=52393),
            cri=cri,
        )
        knxipframe2 = KNXIPFrame.init_from_body(connect_request)

        assert knxipframe2.to_knx() == raw

    def test_from_knx_wrong_length_of_cri(self) -> None:
        """Test parsing and streaming wrong ConnectRequest."""
        raw = bytes.fromhex(
            "06 10 02 05 00 1A 08 01 C0 A8 2A 01 84 95 08 01"
            "C0 A8 2A 01 CC A9 02 04 02 00"
        )
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)

    def test_from_knx_wrong_length_of_cri2(self) -> None:
        """Test parsing and streaming wrong ConnectRequest."""
        raw = bytes.fromhex(
            "06 10 02 05 00 18 08 01 C0 A8 2A 01 84 95 08 01"
            + "C0 A8 2A 01 CC A9 03 03"
        )
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)

    def test_from_knx_wrong_length_of_cri3(self) -> None:
        """Test parsing and streaming wrong ConnectRequest."""
        raw = bytes.fromhex(
            "06 10 02 05 00 18 08 01 C0 A8 2A 01 84 95 08 01"
            + "C0 A8 2A 01 CC A9 01 03"
        )
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)

    def test_from_knx_wrong_length_of_cri4(self) -> None:
        """Test parsing and streaming wrong ConnectRequest."""
        raw = bytes.fromhex(
            "06 10 02 05 00 19 08 01 C0 A8 2A 01 84 95 08 01"
            + "C0 A8 2A 01 CC A9 03 03 01"
        )
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)

    def test_from_knx_wrong_cri(self) -> None:
        """Test parsing and streaming wrong ConnectRequest."""
        raw = bytes.fromhex(
            "06 10 02 05 00 1A 08 01 C0 A8 2A 01 84 95 08 01"
            + "C0 A8 2A 01 CC A9 04 04 02"
        )
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)



"""Implementation of Basic KNX 8-Byte signed (2's complement) values."""

from __future__ import annotations

from .dpt import DPTNumeric, DPTStructIntMixin


class DPT8ByteSigned(DPTStructIntMixin, DPTNumeric):
    """
    Abstraction for KNX 8 Byte "64-bit signed".

    DPT 29.***
    """

    dpt_main_number = 29
    dpt_sub_number: int | None = None
    payload_length = 8
    value_type = "8byte_signed"

    value_min = -9_223_372_036_854_775_808
    value_max = 9_223_372_036_854_775_807
    resolution = 1

    _struct_format = ">q"


class DPTActiveEnergy8Byte(DPT8ByteSigned):
    """DPT 29.010 DPT_Active_Energy_V64."""

    dpt_main_number = 29
    dpt_sub_number = 10
    value_type = "active_energy_8byte"
    unit = "Wh"
    ha_device_class = "energy"


class DPTApparantEnergy8Byte(DPT8ByteSigned):
    """DPT 29.011 DPT_Apparant_Energy_V64 (VAh)."""

    dpt_main_number = 29
    dpt_sub_number = 11
    value_type = "apparant_energy_8byte"
    unit = "VAh"


class DPTReactiveEnergy8Byte(DPT8ByteSigned):
    """DPT 29.012 DPT_Reactive_Energy_V64 (VARh)."""

    dpt_main_number = 29
    dpt_sub_number = 12
    value_type = "reactive_energy_8byte"
    unit = "VARh"



"""Unit test for KNX/IP DisconnectResponse objects."""

import pytest

from xknx.exceptions import CouldNotParseKNXIP
from xknx.knxip import DisconnectResponse, ErrorCode, KNXIPFrame


class TestKNXIPDisconnectResponse:
    """Test class for KNX/IP DisconnectResponse objects."""

    def test_disconnect_response(self) -> None:
        """Test parsing and streaming DisconnectResponse KNX/IP packet."""
        raw = bytes((0x06, 0x10, 0x02, 0x0A, 0x00, 0x08, 0x15, 0x25))
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, DisconnectResponse)

        assert knxipframe.body.communication_channel_id == 21
        assert knxipframe.body.status_code == ErrorCode.E_NO_MORE_UNIQUE_CONNECTIONS

        disconnect_response = DisconnectResponse(
            communication_channel_id=21,
            status_code=ErrorCode.E_NO_MORE_UNIQUE_CONNECTIONS,
        )
        knxipframe2 = KNXIPFrame.init_from_body(disconnect_response)

        assert knxipframe2.to_knx() == raw

    def test_from_knx_wrong_length(self) -> None:
        """Test parsing and streaming wrong DisconnectResponse."""
        raw = bytes((0x06, 0x10, 0x02, 0x0A, 0x00, 0x08, 0x15))
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)



"""Connection States from XKNX."""

from enum import Enum


class XknxConnectionState(Enum):
    """Possible connection state values."""

    CONNECTED = "CONNECTED"
    CONNECTING = "CONNECTING"
    DISCONNECTED = "DISCONNECTED"


class XknxConnectionType(Enum):
    """Possible connection type values."""

    NOT_CONNECTED = None
    ROUTING = "Routing"
    ROUTING_SECURE = "Routing IP Secure"
    TUNNEL_SECURE = "Tunnel IP Secure"
    TUNNEL_TCP = "Tunnel TCP"
    TUNNEL_UDP = "Tunnel UDP"



"""Unit test for KNX date objects."""

import datetime
from typing import Any

import pytest

from xknx.dpt import DPTArray, DPTDate
from xknx.dpt.dpt_11 import KNXDate
from xknx.exceptions import ConversionError, CouldNotParseTelegram


class TestKNXDate:
    """Test class for KNX date objects."""

    @pytest.mark.parametrize(
        ("data", "value"),
        [
            ({"year": 1990, "month": 1, "day": 1}, KNXDate(1990, 1, 1)),
            ({"year": 2024, "month": 7, "day": 26}, KNXDate(2024, 7, 26)),
            ({"year": 2089, "month": 12, "day": 31}, KNXDate(2089, 12, 31)),
        ],
    )
    def test_dict(self, data: dict[str, Any], value: KNXDate) -> None:
        """Test from_dict and as_dict methods."""
        assert KNXDate.from_dict(data) == value
        assert value.as_dict() == data

    @pytest.mark.parametrize(
        "data",
        [
            # invalid data
            {"year": 1},
            {"year": "a"},
            {"month": 2, "day": 3},
            {"year": 2, "day": 3},
            {"year": 1, "month": 2},
            {"year": 1, "month": 2, "day": "a"},
            {"year": 1, "month": None, "day": 3},
        ],
    )
    def test_dict_invalid(self, data: dict[str, Any]) -> None:
        """Test from_dict and as_dict methods."""
        with pytest.raises(ValueError):
            KNXDate.from_dict(data)

    @pytest.mark.parametrize(
        ("date", "value"),
        [
            (datetime.date(1990, 1, 1), KNXDate(1990, 1, 1)),
            (datetime.date(2024, 7, 26), KNXDate(2024, 7, 26)),
            (datetime.date(2089, 12, 31), KNXDate(2089, 12, 31)),
        ],
    )
    def test_as_date(self, date: datetime.date, value: KNXDate) -> None:
        """Test from_time and as_time methods."""
        assert KNXDate.from_date(date) == value
        assert value.as_date() == date


class TestDPTDate:
    """Test class for KNX date objects."""

    @pytest.mark.parametrize(
        ("value", "raw"),
        [
            (KNXDate(2002, 1, 4), (0x04, 0x01, 0x02)),
            (KNXDate(1990, 1, 31), (0x1F, 0x01, 0x5A)),
            (KNXDate(2089, 12, 4), (0x04, 0x0C, 0x59)),
        ],
    )
    def test_from_knx(self, value: KNXDate, raw: tuple[int, ...]) -> None:
        """Test parsing and streaming."""
        knx_value = DPTDate.to_knx(value)
        assert knx_value == DPTArray(raw)
        assert DPTDate.from_knx(knx_value) == value

    def test_from_knx_wrong_parameter(self) -> None:
        """Test parsing from DPTDate object from wrong binary values."""
        with pytest.raises(CouldNotParseTelegram):
            DPTDate.from_knx(DPTArray((0xF8, 0x23)))

    def test_to_knx_wrong_value(self) -> None:
        """Test parsing from DPTDate object from wrong string value."""
        with pytest.raises(ConversionError):
            DPTDate.to_knx(KNXDate(2090, 1, 1))  # year out of range
        with pytest.raises(ConversionError):
            DPTDate.to_knx(KNXDate(1990, 0, 1))  # month out of range
        with pytest.raises(ConversionError):
            DPTDate.to_knx(KNXDate(1990, 1, 32))  # day out of range
        with pytest.raises(ConversionError):
            DPTDate.to_knx("hello")

    def test_from_knx_wrong_range_month(self) -> None:
        """Test Exception when parsing DPTDAte from KNX with wrong month."""
        with pytest.raises(ConversionError):
            DPTDate.from_knx(DPTArray((0x04, 0x00, 0x59)))

    def test_from_knx_wrong_range_year(self) -> None:
        """Test Exception when parsing DPTDate from KNX with wrong year."""
        with pytest.raises(ConversionError):
            DPTDate.from_knx(DPTArray((0x04, 0x01, 0x64)))



"""Implementation of KNX DPT 18 Scene control."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any

from .dpt import DPTComplex, DPTComplexData
from .dpt_17 import DPTSceneNumber
from .payload import DPTArray, DPTBinary


@dataclass(slots=True)
class SceneControl(DPTComplexData):
    """Class for scene control."""

    scene_number: int
    learn: bool = False

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> SceneControl:
        """Init from a dictionary."""
        try:
            scene_number = int(data["scene_number"])
            learn = data.get("learn", False)
        except (KeyError, TypeError, ValueError) as err:
            raise ValueError(f"Invalid value for SceneControl: {err}") from err
        if learn not in (True, False):
            raise ValueError(f"Invalid value for SceneControl value `learn`: {learn}")
        return cls(scene_number=scene_number, learn=bool(learn))

    def as_dict(self) -> dict[str, int | str]:
        """Create a JSON serializable dictionary."""
        return {
            "scene_number": self.scene_number,
            "learn": self.learn,
        }


class DPTSceneControl(DPTComplex[SceneControl]):
    """
    Abstraction for KNX 1 Octet Scene Control.

    DPT 18.001
    """

    data_type = SceneControl
    payload_type = DPTArray
    payload_length = 1
    dpt_main_number = 18
    dpt_sub_number = 1
    value_type = "scene_control"

    @classmethod
    def from_knx(cls, payload: DPTArray | DPTBinary) -> SceneControl:
        """Parse/deserialize from KNX/IP raw data."""
        raw = cls.validate_payload(payload)[0]
        return SceneControl(
            learn=bool(raw & 0x80),
            scene_number=(raw & 0x3F) + 1,
        )

    @classmethod
    def _to_knx(cls, value: SceneControl) -> DPTArray:
        """Serialize to KNX/IP raw data."""
        scene_payload = DPTSceneNumber.to_knx(value.scene_number)
        if value.learn:
            return DPTArray(scene_payload.value[0] | 0x80)
        return scene_payload



"""Test for KNX/IP Tunnelling connections."""

import asyncio
from copy import deepcopy
from unittest.mock import AsyncMock, MagicMock, Mock, call, patch

import pytest

from xknx import XKNX
from xknx.cemi import CEMIFrame, CEMILData, CEMIMessageCode
from xknx.dpt import DPTArray
from xknx.exceptions import CommunicationError
from xknx.io import TCPTunnel, UDPTunnel
from xknx.io.const import CONNECTIONSTATE_REQUEST_TIMEOUT, HEARTBEAT_RATE
from xknx.io.gateway_scanner import GatewayDescriptor
from xknx.knxip import (
    HPAI,
    ConnectionStateRequest,
    ConnectionStateResponse,
    ConnectRequest,
    ConnectResponse,
    ConnectResponseData,
    DescriptionRequest,
    DescriptionResponse,
    DisconnectRequest,
    DisconnectResponse,
    ErrorCode,
    KNXIPFrame,
    TunnellingAck,
    TunnellingRequest,
)
from xknx.knxip.knxip_enum import HostProtocol
from xknx.telegram import (
    GroupAddress,
    IndividualAddress,
    Telegram,
    TelegramDirection,
    tpci,
)
from xknx.telegram.apci import GroupValueWrite

from ..conftest import EventLoopClockAdvancer


class TestUDPTunnel:
    """Test class for xknx/io/Tunnel objects."""

    def setup_method(self) -> None:
        """Set up test class."""
        # pylint: disable=attribute-defined-outside-init
        self.xknx = XKNX()
        self.cemi_received_mock = Mock()
        self.tunnel = UDPTunnel(
            self.xknx,
            gateway_ip="192.168.1.2",
            gateway_port=3671,
            local_ip="192.168.1.1",
            local_port=0,
            cemi_received_callback=self.cemi_received_mock,
            auto_reconnect=False,  # for easier testing
            auto_reconnect_wait=3,
            route_back=False,
        )

    @pytest.mark.parametrize(
        "raw",
        [
            # L_Data.ind GroupValueWrite from 1.1.22 to to 5/1/22 with DPT9 payload 0C 3F
            # communication_channel_id: 0x02   sequence_counter: 0x21
            bytes.fromhex("0610 0420 0017 04 02 21 00 2900bcd011162916030080 0c 3f"),
            # L_Data.ind T_Connect from 1.0.250 to 1.0.255 (xknx tunnel endpoint)
            # communication_channel_id: 0x02   sequence_counter: 0x21
            bytes.fromhex("0610 0420 0014 04 02 21 00 2900b06010fa10ff0080"),
            # <UnsupportedCEMIMessage description="CEMI too small. Length: 9; CEMI: 2900b06010fa10ff00" />
            # communication_channel_id: 0x02   sequence_counter: 0x21
            bytes.fromhex("0610 0420 0013 04 02 21 00 2900b06010fa10ff00"),
        ],
    )
    @patch("xknx.io.UDPTunnel._send_tunnelling_ack")
    async def test_tunnel_request_received(
        self, send_ack_mock: Mock, raw: bytes
    ) -> None:
        """Test Tunnel for calling send_ack on frames."""
        raw_cemi = raw[10:]
        self.tunnel.expected_sequence_number = 0x21

        self.tunnel.transport.data_received_callback(raw, ("192.168.1.2", 3671))
        await asyncio.sleep(0)
        self.cemi_received_mock.assert_called_once_with(raw_cemi)
        send_ack_mock.assert_called_once_with(raw[7], raw[8])

    @patch("xknx.io.UDPTunnel._send_tunnelling_ack")
    @patch("xknx.io.UDPTunnel.send_cemi")
    async def test_tunnel_request_received_callback(
        self,
        send_cemi_mock: Mock,
        send_ack_mock: Mock,
    ) -> None:
        """Test Tunnel for responding to point-to-point connection."""
        self.tunnel.cemi_received_callback = self.xknx.knxip_interface.cemi_received
        self.xknx.knxip_interface._interface = self.tunnel
        # set current address so management telegram is processed
        self.xknx.current_address = IndividualAddress("1.0.255")
        # L_Data.ind T_Connect from 1.0.250 to 1.0.255 (xknx tunnel endpoint) - ETS Line-Scan
        # communication_channel_id: 0x02   sequence_counter: 0x81
        raw_ind = bytes.fromhex("0610 0420 0014 04 02 81 00 2900b06010fa10ff0080")

        test_cemi = CEMIFrame.from_knx(raw_ind[10:])
        test_telegram = test_cemi.data.telegram()
        test_telegram.direction = TelegramDirection.INCOMING
        self.tunnel.expected_sequence_number = 0x81

        response_telegram = Telegram(
            destination_address=IndividualAddress(test_telegram.source_address),
            tpci=tpci.TDisconnect(),
        )

        self.tunnel.transport.data_received_callback(raw_ind, ("192.168.1.2", 3671))
        await asyncio.sleep(0)
        response_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_REQ,
            data=CEMILData.init_from_telegram(
                response_telegram,
                src_addr=IndividualAddress("1.0.255"),
            ),
        )
        assert send_cemi_mock.call_args_list == [
            call(response_cemi),
        ]
        send_ack_mock.assert_called_once_with(raw_ind[7], raw_ind[8])

    @patch("xknx.io.transport.udp_transport.UDPTransport.send")
    @patch("xknx.io.transport.udp_transport.UDPTransport.stop")
    async def test_repeated_tunnel_request(
        self,
        mock_transport_stop: Mock,
        mock_transport_send: Mock,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test Tunnel for receiving repeated TunnellingRequest frames."""
        self.tunnel.transport.transport = Mock()
        self.tunnel.communication_channel = 1
        self.tunnel.expected_sequence_number = 10

        test_telegram = Telegram(
            destination_address=GroupAddress(1),
            payload=GroupValueWrite(DPTArray((1,))),
        )
        cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_IND,
            data=CEMILData.init_from_telegram(test_telegram),
        )
        test_frame = KNXIPFrame.init_from_body(
            TunnellingRequest(
                communication_channel_id=1,
                sequence_counter=10,
                raw_cemi=cemi.to_knx(),
            )
        )
        test_ack = KNXIPFrame.init_from_body(TunnellingAck(sequence_counter=10))
        test_frame_9 = deepcopy(test_frame)
        test_frame_9.body.sequence_counter = 9

        # first frame - ACK and processed
        self.tunnel._request_received(test_frame, None, None)
        await time_travel(0)
        assert mock_transport_send.call_args_list == [call(test_ack, addr=None)]
        mock_transport_send.reset_mock()
        assert self.tunnel.expected_sequence_number == 11
        assert self.cemi_received_mock.call_count == 1
        # same sequence number as before - ACK, not processed
        self.tunnel._request_received(test_frame, None, None)
        await time_travel(0)
        assert mock_transport_send.call_args_list == [call(test_ack, addr=None)]
        mock_transport_send.reset_mock()
        assert self.tunnel.expected_sequence_number == 11
        assert self.cemi_received_mock.call_count == 1
        # wrong sequence number - no ACK, not processed
        # reconnect if `auto_reconnect` was True, disconnect if False
        self.tunnel._request_received(test_frame_9, None, None)
        mock_transport_send.assert_not_called()
        mock_transport_stop.assert_not_called()
        # Disconnect after 2 seconds of no valid frame
        await time_travel(2)
        mock_transport_stop.assert_called_once()

        assert self.tunnel.expected_sequence_number == 11
        assert self.cemi_received_mock.call_count == 1

    @patch("xknx.io.transport.udp_transport.UDPTransport.send")
    @patch("xknx.io.transport.udp_transport.UDPTransport.stop")
    async def test_tunnel_send_retry(
        self,
        mock_transport_stop: Mock,
        mock_transport_send: Mock,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test tunnel resends the telegram when no ACK was received."""
        self.tunnel.transport.transport = Mock()
        self.tunnel.communication_channel = 1
        self.tunnel.sequence_number = 23
        self.tunnel.expected_sequence_number = 15

        test_telegram = Telegram(
            destination_address=GroupAddress(1),
            payload=GroupValueWrite(DPTArray((1,))),
        )
        test_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_REQ,
            data=CEMILData.init_from_telegram(
                test_telegram,
                src_addr=self.tunnel._src_address,
            ),
        )
        request = KNXIPFrame.init_from_body(
            TunnellingRequest(
                communication_channel_id=self.tunnel.communication_channel,
                sequence_counter=self.tunnel.sequence_number,
                raw_cemi=test_cemi.to_knx(),
            )
        )
        test_ack = KNXIPFrame.init_from_body(
            TunnellingAck(sequence_counter=self.tunnel.sequence_number)
        )
        confirmation_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_CON,
            data=CEMILData.init_from_telegram(test_telegram),
        )
        confirmation = KNXIPFrame.init_from_body(
            TunnellingRequest(
                communication_channel_id=1,
                sequence_counter=15,
                raw_cemi=confirmation_cemi.to_knx(),
            )
        )
        confirmation_ack = KNXIPFrame.init_from_body(TunnellingAck(sequence_counter=15))

        task = asyncio.create_task(self.tunnel.send_cemi(test_cemi))
        await time_travel(0)
        assert mock_transport_send.call_args_list == [call(request, addr=None)]
        mock_transport_send.reset_mock()
        # no ACK received, resend same telegram
        await time_travel(1)
        assert mock_transport_send.call_args_list == [call(request, addr=None)]
        mock_transport_send.reset_mock()
        self.tunnel.transport.handle_knxipframe(test_ack, HPAI())
        await time_travel(0)
        assert task.done()
        await task
        # L_Data.con ACK for UDP tunneling
        self.tunnel.transport.handle_knxipframe(confirmation, HPAI())
        await time_travel(0)
        assert mock_transport_send.call_args_list == [call(confirmation_ack, addr=None)]
        mock_transport_send.reset_mock()

        # Test raise after 2 missed ACKs (`auto_reconnect` is False)
        task = asyncio.create_task(self.tunnel.send_cemi(test_cemi))
        # no ACKs received, for 2x wait time (with advancing the loop in between)
        await time_travel(1)
        await time_travel(1)
        # 2x TunnellingRequest  + 1x DisconnectRequest
        assert mock_transport_send.call_count == 3
        mock_transport_send.reset_mock()
        mock_transport_stop.assert_called_once()
        with pytest.raises(CommunicationError):
            await task

    @pytest.mark.parametrize(
        ("route_back", "data_endpoint_addr", "local_endpoint"),
        [
            (False, ("192.168.1.2", 56789), HPAI("192.168.1.1", 12345)),
            (True, None, HPAI()),
        ],
    )
    @patch("xknx.io.transport.udp_transport.UDPTransport.send")
    @patch("xknx.io.transport.udp_transport.UDPTransport.stop")
    @patch("xknx.io.transport.udp_transport.UDPTransport.getsockname")
    @patch("xknx.io.transport.udp_transport.UDPTransport.connect")
    async def test_tunnel_connect_send_disconnect(
        self,
        mock_transport_connect: Mock,
        mock_transport_getsockname: Mock,
        mock_transport_stop: Mock,
        mock_transport_send: Mock,
        time_travel: EventLoopClockAdvancer,
        route_back: bool,
        data_endpoint_addr: tuple[str, int] | None,
        local_endpoint: HPAI,
    ) -> None:
        """Test initiating a tunnelling connection."""
        local_addr = ("192.168.1.1", 12345)
        remote_addr = ("192.168.1.2", 3671)
        self.tunnel.route_back = route_back
        gateway_data_endpoint = (
            HPAI(*data_endpoint_addr) if data_endpoint_addr else HPAI()
        )
        mock_transport_getsockname.return_value = local_addr

        # Connect
        connect_request = ConnectRequest(
            control_endpoint=local_endpoint,
            data_endpoint=local_endpoint,
        )
        connect_frame = KNXIPFrame.init_from_body(connect_request)

        connection_task = asyncio.create_task(self.tunnel.connect())
        await time_travel(0)
        mock_transport_connect.assert_called_once()
        mock_transport_send.assert_called_once_with(connect_frame)

        connect_response_frame = KNXIPFrame.init_from_body(
            ConnectResponse(
                communication_channel=23,
                data_endpoint=gateway_data_endpoint,
                crd=ConnectResponseData(individual_address=IndividualAddress(7)),
            )
        )
        self.tunnel.transport.handle_knxipframe(
            connect_response_frame, HPAI(*remote_addr)
        )
        await connection_task
        assert self.tunnel._data_endpoint_addr == data_endpoint_addr
        assert self.tunnel._src_address == IndividualAddress(7)

        # Send - use data endpoint
        mock_transport_send.reset_mock()
        test_telegram = Telegram(
            destination_address=GroupAddress(1),
            payload=GroupValueWrite(DPTArray((1,))),
        )
        test_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_REQ,
            data=CEMILData.init_from_telegram(
                test_telegram,
                src_addr=IndividualAddress(7),
            ),
        )
        test_telegram_frame = KNXIPFrame.init_from_body(
            TunnellingRequest(
                communication_channel_id=23,
                sequence_counter=0,
                raw_cemi=test_cemi.to_knx(),
            )
        )
        send_task = asyncio.create_task(self.tunnel.send_cemi(test_cemi))
        await time_travel(0)
        mock_transport_send.assert_called_once_with(
            test_telegram_frame, addr=data_endpoint_addr
        )
        # skip ack and confirmation
        assert not send_task.done()

        # Disconnect
        mock_transport_send.reset_mock()
        disconnect_request = DisconnectRequest(
            communication_channel_id=23,
            control_endpoint=local_endpoint,
        )
        disconnect_frame = KNXIPFrame.init_from_body(disconnect_request)

        disconnection_task = asyncio.create_task(self.tunnel.disconnect())
        await time_travel(0)
        mock_transport_send.assert_called_once_with(disconnect_frame)

        disconnect_response_frame = KNXIPFrame.init_from_body(
            DisconnectResponse(communication_channel_id=23)
        )
        self.tunnel.transport.handle_knxipframe(
            disconnect_response_frame, HPAI(*remote_addr)
        )
        await disconnection_task
        mock_transport_stop.assert_called_once()

    @patch("xknx.io.transport.udp_transport.UDPTransport.send")
    @patch("xknx.io.transport.udp_transport.UDPTransport.stop")
    @patch("xknx.io.transport.udp_transport.UDPTransport.getsockname")
    @patch("xknx.io.transport.udp_transport.UDPTransport.connect")
    # no need to wait for ConnectResponse in this test, so mock connect request
    @patch("xknx.io.tunnel.UDPTunnel._connect_request")
    async def test_tunnel_reconnect(
        self,
        mock_connect_request: Mock,
        mock_transport_connect: Mock,
        mock_transport_getsockname: Mock,
        mock_transport_stop: Mock,
        mock_transport_send: Mock,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test tunnel reconnection."""
        local_addr = ("192.168.1.1", 12345)
        remote_addr = ("192.168.1.2", 3671)

        # prepare and initialize transport
        def transport_connect_success_side_effect() -> None:
            self.tunnel.transport.transport = Mock()

        def transport_connect_error_side_effect() -> None:
            raise CommunicationError("test")

        def transport_stop_side_effect() -> None:
            self.tunnel.transport.transport = None

        mock_transport_connect.side_effect = transport_connect_error_side_effect
        mock_transport_getsockname.return_value = local_addr
        mock_transport_stop.side_effect = transport_stop_side_effect
        transport_connect_success_side_effect()

        # prepare and initialize tunnel
        self.tunnel.auto_reconnect = True
        self.tunnel.communication_channel = 1

        # server sends disconnect request
        disconnect_request_frame = KNXIPFrame.init_from_body(
            DisconnectRequest(
                communication_channel_id=1, control_endpoint=HPAI(*remote_addr)
            )
        )
        self.tunnel.transport.handle_knxipframe(
            disconnect_request_frame, HPAI(*remote_addr)
        )
        disconnect_response_frame = KNXIPFrame.init_from_body(
            DisconnectResponse(communication_channel_id=1)
        )
        mock_transport_send.assert_called_once_with(disconnect_response_frame)
        assert self.tunnel.communication_channel is None

        assert mock_transport_stop.call_count == 0
        # reconnect task will start next loop iteration
        await asyncio.sleep(0)
        # once for initial stop before reconnect and once after first
        # reconnect attempt fails due to CommunicationError
        assert mock_transport_stop.call_count == 2
        # first reconnect attempt sent immediately... fails due to side effect
        mock_transport_connect.assert_called_once()
        mock_transport_connect.reset_mock()
        assert not self.tunnel._reconnect_task.done()

        # second reconnect attempt after wait time
        mock_transport_connect.side_effect = transport_connect_success_side_effect
        await time_travel(self.tunnel.auto_reconnect_wait)
        mock_transport_connect.assert_called_once()
        assert self.tunnel._reconnect_task is None

    @patch("xknx.io.transport.udp_transport.UDPTransport.send")
    @patch("xknx.io.transport.udp_transport.UDPTransport.getsockname")
    async def test_tunnel_request_description(
        self,
        mock_transport_getsockname: Mock,
        mock_transport_send: Mock,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test tunnel requesting and returning description of connected interface."""
        local_addr = ("192.168.1.1", 12345)
        mock_transport_getsockname.return_value = local_addr

        description_request = KNXIPFrame.init_from_body(
            DescriptionRequest(control_endpoint=self.tunnel.local_hpai)
        )
        description_response = KNXIPFrame.init_from_body(DescriptionResponse())

        task = asyncio.create_task(self.tunnel.request_description())
        await time_travel(0)
        mock_transport_send.assert_called_once_with(description_request)
        self.tunnel.transport.handle_knxipframe(description_response, HPAI())
        await time_travel(0)
        assert task.done()
        assert isinstance(task.result(), GatewayDescriptor)
        assert mock_transport_send.call_count == 1
        await task


class TestTCPTunnel:
    """Test class for xknx/io/TCPTunnel objects."""

    def setup_method(self) -> None:
        """Set up test class."""
        # pylint: disable=attribute-defined-outside-init
        self.xknx = XKNX()
        self.cemi_received_mock = AsyncMock()
        self.tunnel = TCPTunnel(
            self.xknx,
            gateway_ip="192.168.1.2",
            gateway_port=3671,
            cemi_received_callback=self.cemi_received_mock,
            auto_reconnect=False,
            auto_reconnect_wait=3,
        )

    @patch("xknx.io.transport.tcp_transport.TCPTransport.connect")
    @patch("xknx.io.transport.tcp_transport.TCPTransport.getsockname")
    @patch("xknx.io.transport.tcp_transport.TCPTransport.send")
    @patch("xknx.io.transport.tcp_transport.TCPTransport.stop")
    @patch("xknx.io.tunnel.TCPTunnel._tunnel_lost")
    async def test_tunnel_heartbeat(
        self,
        mock_tunnel_lost: MagicMock,
        mock_transport_stop: MagicMock,
        mock_transport_send: MagicMock,
        mock_transport_getsockname: MagicMock,
        mock_transport_connect: MagicMock,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test tunnel sends heartbeat frame."""
        local_addr = ("192.168.1.1", 12345)
        remote_hpai = HPAI(
            ip_addr="192.168.1.2", port=3671, protocol=HostProtocol.IPV4_TCP
        )
        mock_transport_getsockname.return_value = local_addr

        # Connect
        connection_task = asyncio.create_task(self.tunnel.connect())
        await time_travel(0)
        mock_transport_connect.assert_called_once()

        connect_response_frame = KNXIPFrame.init_from_body(
            ConnectResponse(
                communication_channel=23,
                data_endpoint=HPAI(protocol=HostProtocol.IPV4_TCP),
                crd=ConnectResponseData(individual_address=IndividualAddress(7)),
            )
        )
        self.tunnel.transport.handle_knxipframe(connect_response_frame, remote_hpai)
        await connection_task
        mock_transport_send.reset_mock()

        # Send heartbeat
        heartbeat_request = KNXIPFrame.init_from_body(
            ConnectionStateRequest(
                communication_channel_id=23,
                control_endpoint=HPAI(protocol=HostProtocol.IPV4_TCP),
            )
        )
        heartbeat_response = KNXIPFrame.init_from_body(
            ConnectionStateResponse(
                communication_channel_id=23,
                status_code=ErrorCode.E_NO_ERROR,
            )
        )
        await time_travel(HEARTBEAT_RATE)
        mock_transport_send.assert_called_once_with(heartbeat_request)
        mock_transport_send.reset_mock()
        self.tunnel.transport.handle_knxipframe(heartbeat_response, remote_hpai)
        # test no retry-heartbeat was sent
        await time_travel(CONNECTIONSTATE_REQUEST_TIMEOUT)
        mock_transport_send.assert_not_called()
        # next regular heartbeat
        await time_travel(HEARTBEAT_RATE - CONNECTIONSTATE_REQUEST_TIMEOUT)
        mock_transport_send.assert_called_once_with(heartbeat_request)
        mock_transport_send.reset_mock()
        mock_tunnel_lost.assert_not_called()

    @patch("xknx.io.transport.tcp_transport.TCPTransport.connect")
    @patch("xknx.io.transport.tcp_transport.TCPTransport.getsockname")
    @patch("xknx.io.transport.tcp_transport.TCPTransport.send")
    @patch("xknx.io.transport.tcp_transport.TCPTransport.stop")
    @patch("xknx.io.tunnel.TCPTunnel._tunnel_lost")
    async def test_tunnel_heartbeat_no_answer(
        self,
        mock_tunnel_lost: MagicMock,
        mock_transport_stop: MagicMock,
        mock_transport_send: MagicMock,
        mock_transport_getsockname: MagicMock,
        mock_transport_connect: MagicMock,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test tunnel sends heartbeat frame."""
        local_addr = ("192.168.1.1", 12345)
        remote_hpai = HPAI(
            ip_addr="192.168.1.2", port=3671, protocol=HostProtocol.IPV4_TCP
        )
        mock_transport_getsockname.return_value = local_addr

        # Connect
        connection_task = asyncio.create_task(self.tunnel.connect())
        await time_travel(0)
        mock_transport_connect.assert_called_once()

        connect_response_frame = KNXIPFrame.init_from_body(
            ConnectResponse(
                communication_channel=23,
                data_endpoint=HPAI(protocol=HostProtocol.IPV4_TCP),
                crd=ConnectResponseData(individual_address=IndividualAddress(7)),
            )
        )
        self.tunnel.transport.handle_knxipframe(connect_response_frame, remote_hpai)
        await connection_task
        mock_transport_send.reset_mock()

        # Send heartbeat
        heartbeat_request = KNXIPFrame.init_from_body(
            ConnectionStateRequest(
                communication_channel_id=23,
                control_endpoint=HPAI(protocol=HostProtocol.IPV4_TCP),
            )
        )
        await time_travel(HEARTBEAT_RATE)
        mock_transport_send.assert_called_once_with(heartbeat_request)
        mock_transport_send.reset_mock()
        # no answer - repeat 3 times
        await time_travel(CONNECTIONSTATE_REQUEST_TIMEOUT)
        mock_transport_send.assert_called_once_with(heartbeat_request)
        mock_transport_send.reset_mock()
        await time_travel(CONNECTIONSTATE_REQUEST_TIMEOUT)
        mock_transport_send.assert_called_once_with(heartbeat_request)
        mock_transport_send.reset_mock()
        await time_travel(CONNECTIONSTATE_REQUEST_TIMEOUT)
        mock_transport_send.assert_called_once_with(heartbeat_request)
        mock_transport_send.reset_mock()
        mock_tunnel_lost.assert_not_called()
        await time_travel(CONNECTIONSTATE_REQUEST_TIMEOUT)
        # no answer - tunnel lost
        mock_tunnel_lost.assert_called_once()

    @patch("xknx.io.transport.tcp_transport.TCPTransport.connect")
    @patch("xknx.io.transport.tcp_transport.TCPTransport.getsockname")
    @patch("xknx.io.transport.tcp_transport.TCPTransport.send")
    @patch("xknx.io.transport.tcp_transport.TCPTransport.stop")
    @patch("xknx.io.tunnel.TCPTunnel._tunnel_lost")
    async def test_tunnel_heartbeat_error(
        self,
        mock_tunnel_lost: MagicMock,
        mock_transport_stop: MagicMock,
        mock_transport_send: MagicMock,
        mock_transport_getsockname: MagicMock,
        mock_transport_connect: MagicMock,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test tunnel sends heartbeat frame."""
        local_addr = ("192.168.1.1", 12345)
        remote_hpai = HPAI(
            ip_addr="192.168.1.2", port=3671, protocol=HostProtocol.IPV4_TCP
        )
        mock_transport_getsockname.return_value = local_addr

        # Connect
        connection_task = asyncio.create_task(self.tunnel.connect())
        await time_travel(0)
        mock_transport_connect.assert_called_once()

        connect_response_frame = KNXIPFrame.init_from_body(
            ConnectResponse(
                communication_channel=23,
                data_endpoint=HPAI(protocol=HostProtocol.IPV4_TCP),
                crd=ConnectResponseData(individual_address=IndividualAddress(7)),
            )
        )
        self.tunnel.transport.handle_knxipframe(connect_response_frame, remote_hpai)
        await connection_task
        mock_transport_send.reset_mock()

        # Send heartbeat
        heartbeat_request = KNXIPFrame.init_from_body(
            ConnectionStateRequest(
                communication_channel_id=23,
                control_endpoint=HPAI(protocol=HostProtocol.IPV4_TCP),
            )
        )
        heartbeat_error_response = KNXIPFrame.init_from_body(
            ConnectionStateResponse(
                communication_channel_id=23,
                status_code=ErrorCode.E_CONNECTION_ID,
            )
        )
        await time_travel(HEARTBEAT_RATE)
        mock_transport_send.assert_called_once_with(heartbeat_request)
        mock_transport_send.reset_mock()
        self.tunnel.transport.handle_knxipframe(heartbeat_error_response, remote_hpai)
        # repeat 3 times
        await time_travel(0)
        mock_transport_send.assert_called_once_with(heartbeat_request)
        mock_transport_send.reset_mock()
        self.tunnel.transport.handle_knxipframe(heartbeat_error_response, remote_hpai)
        await time_travel(0)
        mock_transport_send.assert_called_once_with(heartbeat_request)
        mock_transport_send.reset_mock()
        self.tunnel.transport.handle_knxipframe(heartbeat_error_response, remote_hpai)
        await time_travel(0)
        mock_transport_send.assert_called_once_with(heartbeat_request)
        mock_transport_send.reset_mock()
        mock_tunnel_lost.assert_not_called()
        self.tunnel.transport.handle_knxipframe(heartbeat_error_response, remote_hpai)
        await time_travel(0)
        # third retry had an error - tunnel lost
        mock_tunnel_lost.assert_called_once()



"""Unit test for KNX/IP DIB objects."""

import pytest

from xknx.exceptions import CouldNotParseKNXIP
from xknx.knxip import (
    DIB,
    DIBDeviceInformation,
    DIBGeneric,
    DIBSecuredServiceFamilies,
    DIBServiceFamily,
    DIBSuppSVCFamilies,
    DIBTunnelingInfo,
    DIBTypeCode,
    KNXMedium,
)
from xknx.telegram import IndividualAddress


class TestKNXIPDIB:
    """Test class for KNX/IP DIB objects."""

    def test_dib_base(self) -> None:
        """Test parsing and streaming KNX/IP DIB packet."""
        raw = bytes(
            (0x0C, 0x02, 0x02, 0x01, 0x03, 0x02, 0x04, 0x01, 0x05, 0x01, 0x07, 0x01)
        )
        dib = DIBGeneric()
        assert dib.from_knx(raw) == 12
        assert dib.dtc == DIBTypeCode.SUPP_SVC_FAMILIES
        assert dib.to_knx() == raw
        assert dib.calculated_length() == 12

    def test_dib_wrong_input(self) -> None:
        """Test parsing of wrong KNX/IP DIB packet."""
        raw = (0x08, 0x01, 0xC0, 0xA8, 0x2A)
        with pytest.raises(CouldNotParseKNXIP):
            DIBGeneric().from_knx(raw)

    def test_device_info(self) -> None:
        """Test parsing of device info."""
        raw = bytes.fromhex(
            "36 01 02 00 11 00 23 42 13 37 13 37 13 37 E0 00"
            "17 0C 00 01 02 03 04 05 47 69 72 61 20 4B 4E 58"
            "2F 49 50 2D 52 6F 75 74 65 72 00 00 00 00 00 00"
            "00 00 00 00 00 00"
        )

        dib = DIB.determine_dib(raw)
        assert isinstance(dib, DIBDeviceInformation)
        assert dib.from_knx(raw) == DIBDeviceInformation.LENGTH
        assert dib.knx_medium == KNXMedium.TP1
        assert dib.programming_mode is False
        assert dib.individual_address == IndividualAddress("1.1.0")
        assert dib.name == "Gira KNX/IP-Router"
        assert dib.mac_address == "00:01:02:03:04:05"
        assert dib.multicast_address == "224.0.23.12"
        assert dib.serial_number == "13:37:13:37:13:37"
        assert dib.project_number == 564
        assert dib.installation_number == 2
        assert dib.to_knx() == raw

    def test_dib_sup_svc_families_router(self) -> None:
        """Test parsing of svc families."""
        raw = bytes(
            (0x0C, 0x02, 0x02, 0x01, 0x03, 0x02, 0x04, 0x01, 0x05, 0x01, 0x07, 0x01)
        )

        dib = DIB.determine_dib(raw)
        assert isinstance(dib, DIBSuppSVCFamilies)
        assert dib.from_knx(raw) == 12

        assert dib.families == [
            DIBSuppSVCFamilies.Family(DIBServiceFamily.CORE, 1),
            DIBSuppSVCFamilies.Family(DIBServiceFamily.DEVICE_MANAGEMENT, 2),
            DIBSuppSVCFamilies.Family(DIBServiceFamily.TUNNELING, 1),
            DIBSuppSVCFamilies.Family(DIBServiceFamily.ROUTING, 1),
            DIBSuppSVCFamilies.Family(
                DIBServiceFamily.REMOTE_CONFIGURATION_DIAGNOSIS, 1
            ),
        ]

        assert dib.to_knx() == raw

        assert dib.supports(DIBServiceFamily.CORE)
        assert dib.supports(DIBServiceFamily.DEVICE_MANAGEMENT)
        assert dib.supports(DIBServiceFamily.DEVICE_MANAGEMENT, version=2)
        assert dib.supports(DIBServiceFamily.TUNNELING)
        assert not dib.supports(DIBServiceFamily.TUNNELING, version=2)
        assert dib.supports(DIBServiceFamily.ROUTING, version=1)

        assert dib.version(DIBServiceFamily.CORE) == 1
        assert dib.version(DIBServiceFamily.DEVICE_MANAGEMENT) == 2
        assert dib.version(DIBServiceFamily.TUNNELING) == 1
        assert dib.version(DIBServiceFamily.ROUTING) == 1

    def test_dib_sup_svc_families_interface(self) -> None:
        """Test parsing of svc families."""
        raw = bytes((0x0A, 0x02, 0x02, 0x02, 0x03, 0x02, 0x04, 0x02, 0x07, 0x01))

        dib = DIB.determine_dib(raw)
        assert isinstance(dib, DIBSuppSVCFamilies)
        assert dib.from_knx(raw) == 10

        assert dib.families == [
            DIBSuppSVCFamilies.Family(DIBServiceFamily.CORE, 2),
            DIBSuppSVCFamilies.Family(DIBServiceFamily.DEVICE_MANAGEMENT, 2),
            DIBSuppSVCFamilies.Family(DIBServiceFamily.TUNNELING, 2),
            DIBSuppSVCFamilies.Family(
                DIBServiceFamily.REMOTE_CONFIGURATION_DIAGNOSIS, 1
            ),
        ]

        assert dib.to_knx() == raw

        assert dib.supports(DIBServiceFamily.TUNNELING)
        assert dib.supports(DIBServiceFamily.TUNNELING, version=2)
        assert not dib.supports(DIBServiceFamily.ROUTING)
        assert not dib.supports(DIBServiceFamily.ROUTING, version=2)

        assert dib.version(DIBServiceFamily.CORE) == 2
        assert dib.version(DIBServiceFamily.DEVICE_MANAGEMENT) == 2
        assert dib.version(DIBServiceFamily.TUNNELING) == 2
        assert dib.version(DIBServiceFamily.ROUTING) is None

    def test_dib_secured_service_families(self) -> None:
        """Test parsing of secured service families."""
        raw = bytes((0x08, 0x06, 0x03, 0x01, 0x04, 0x01, 0x05, 0x01))

        dib = DIB.determine_dib(raw)
        assert isinstance(dib, DIBSecuredServiceFamilies)
        assert dib.from_knx(raw) == 8

        assert dib.families == [
            DIBSuppSVCFamilies.Family(DIBServiceFamily.DEVICE_MANAGEMENT, 1),
            DIBSuppSVCFamilies.Family(DIBServiceFamily.TUNNELING, 1),
            DIBSuppSVCFamilies.Family(DIBServiceFamily.ROUTING, 1),
        ]

        assert dib.to_knx() == raw

        assert dib.supports(DIBServiceFamily.TUNNELING)
        assert dib.supports(DIBServiceFamily.TUNNELING, version=1)
        assert dib.supports(DIBServiceFamily.ROUTING)
        assert not dib.supports(DIBServiceFamily.ROUTING, version=2)
        assert dib.supports(DIBServiceFamily.DEVICE_MANAGEMENT)

    def test_dib_tunneling_info(self) -> None:
        """Test parsing of tunneling info."""
        raw = (
            b"\x24\x07\x00\xf8\x40\x01\x00\x05\x40\x02\x00\x05\x40\x03\x00\x05"
            b"\x40\x04\x00\x05\x40\x05\x00\x05\x40\x06\x00\x05\x40\x07\x00\x05"
            b"\x40\x08\x00\x06"
        )

        dib = DIB.determine_dib(raw)
        assert isinstance(dib, DIBTunnelingInfo)
        assert dib.from_knx(raw) == 36

        assert dib.max_apdu_length == 248

        assert len(dib.slots) == 8
        for address in ["4.0.1", "4.0.2", "4.0.3", "4.0.4", "4.0.5", "4.0.6", "4.0.7"]:
            assert dib.slots[IndividualAddress(address)].usable
            assert not dib.slots[IndividualAddress(address)].authorized
            assert dib.slots[IndividualAddress(address)].free
        assert dib.slots[IndividualAddress("4.0.8")].usable
        assert dib.slots[IndividualAddress("4.0.8")].authorized
        assert not dib.slots[IndividualAddress("4.0.8")].free

        assert dib.to_knx() == raw



"""
Module for managing an xyY-color remote value.

DPT 242.600.
"""

from __future__ import annotations

from xknx.dpt import DPTColorXYY, XYYColor

from .remote_value import RemoteValue


class RemoteValueColorXYY(RemoteValue[XYYColor]):
    """Abstraction for remote value of KNX DPT 242.600 (DPT_Colour_xyY)."""

    dpt_class = DPTColorXYY



"""Unit test for KNX/IP Tunnelling Request/Response."""

import asyncio
from unittest.mock import Mock, patch

from xknx.io.self_description import DescriptionQuery, request_description
from xknx.io.transport.udp_transport import UDPTransport
from xknx.knxip import HPAI, DescriptionRequest, KNXIPFrame, SearchRequestExtended
from xknx.telegram import IndividualAddress

from ..conftest import EventLoopClockAdvancer


class TestSelfDescription:
    """Test class for KNX/IP self description."""

    # Core v1 device MDT KNX IP Router SCN-IP100.02
    description_response_v1_raw = bytes.fromhex(
        "06 10 02 04 00 46 36 01 02 00 20 00 00 00 00"
        "83 47 7f 01 24 e0 00 17 0c cc 1b e0 80 04 c4"
        "4d 44 54 20 4b 4e 58 20 49 50 20 52 6f 75 74"
        "65 72 00 00 00 00 00 00 00 00 00 00 00 00 00"
        "0a 02 02 01 03 01 04 01 05 01"
    )
    # Core v2 responses from a KNX IP Router 752 secure
    description_response_raw = bytes.fromhex(
        "06 10 02 04 00 4e 36 01 02 00 50 00 00 00 00"
        "c5 01 04 d5 9d e0 00 17 0d 00 24 6d 02 94 c4"
        "4b 4e 58 20 49 50 20 52 6f 75 74 65 72 20 37"
        "35 32 20 73 65 63 75 72 65 00 00 00 00 00 00"
        "0a 02 02 02 03 02 04 02 05 02 08 08 00 00 00"
        "37 09 1a"
    )
    search_response_extended_raw = bytes.fromhex(
        "06 10 02 0c 00 84 08 01 0a 01 01 5f 0e 57 36"
        "01 02 00 50 00 00 00 00 c5 01 04 d5 9d e0 00"
        "17 0d 00 24 6d 02 94 c4 4b 4e 58 20 49 50 20"
        "52 6f 75 74 65 72 20 37 35 32 20 73 65 63 75"
        "72 65 00 00 00 00 00 00 0c 02 02 02 03 02 04"
        "02 05 02 09 01 08 06 03 01 04 01 05 01 24 07"
        "00 37 50 01 ff fd 50 02 ff fd 50 03 ff fd 50"
        "04 ff fd 50 05 ff fd 50 06 ff fd 50 07 ff fd"
        "50 08 ff fd 08 08 00 00 00 37 09 1a"
    )

    @patch("xknx.io.transport.udp_transport.UDPTransport.send")
    @patch("xknx.io.transport.udp_transport.UDPTransport.getsockname")
    async def test_description_query(
        self,
        mock_transport_getsockname: Mock,
        mock_transport_send: Mock,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test DescriptionQuery class."""
        local_addr = ("127.0.0.1", 12345)
        remote_addr = ("127.0.0.2", 54321)
        transport_mocked = UDPTransport(local_addr=local_addr, remote_addr=remote_addr)
        mock_transport_getsockname.return_value = local_addr

        description_request = KNXIPFrame.init_from_body(
            DescriptionRequest(control_endpoint=HPAI(*local_addr))
        )

        description_query = DescriptionQuery(
            transport=transport_mocked, local_hpai=HPAI(*local_addr)
        )
        task = asyncio.create_task(description_query.start())
        await time_travel(0)
        mock_transport_send.assert_called_once_with(description_request)
        transport_mocked.data_received_callback(
            raw=self.description_response_raw, source=remote_addr
        )
        await task
        assert description_query.gateway_descriptor is not None
        assert description_query.gateway_descriptor.name == "KNX IP Router 752 secure"

    async def test_request_description_v1(
        self,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test request_description function with Core v1 device."""
        local_addr = ("127.0.0.1", 12345)
        remote_addr = ("127.0.0.2", 54321)

        with (
            patch(
                "xknx.io.self_description.UDPTransport.connect"
            ) as transport_connect_mock,
            patch(
                "xknx.io.self_description.UDPTransport.getsockname",
                return_value=local_addr,
            ),
            patch("xknx.io.self_description.UDPTransport.send") as transport_send_mock,
            patch("xknx.io.self_description.UDPTransport.stop") as transport_stop_mock,
            patch(
                "xknx.io.self_description.DescriptionQuery", wraps=DescriptionQuery
            ) as description_query_mock,
        ):
            task = asyncio.create_task(request_description(remote_addr[0]))
            await time_travel(0)
            transport_connect_mock.assert_called_once_with()
            assert transport_send_mock.call_count == 1
            assert isinstance(
                transport_send_mock.call_args[0][0].body, DescriptionRequest
            )
            _transport = description_query_mock.call_args.kwargs["transport"]
            _transport.data_received_callback(
                self.description_response_v1_raw, remote_addr
            )
            gateway = await task
            transport_stop_mock.assert_called_once()
            assert transport_send_mock.call_count == 1

            assert gateway.core_version == 1
            assert gateway.name == "MDT KNX IP Router"
            assert gateway.tunnelling_requires_secure is None
            assert gateway.individual_address == IndividualAddress("2.0.0")

    async def test_request_description_extended(
        self,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test request_description function with Core v2 device."""
        local_addr = ("127.0.0.1", 12345)
        remote_addr = ("127.0.0.2", 54321)

        with (
            patch(
                "xknx.io.self_description.UDPTransport.connect"
            ) as transport_connect_mock,
            patch(
                "xknx.io.self_description.UDPTransport.getsockname",
                return_value=local_addr,
            ),
            patch("xknx.io.self_description.UDPTransport.send") as transport_send_mock,
            patch("xknx.io.self_description.UDPTransport.stop") as transport_stop_mock,
            patch(
                "xknx.io.self_description.DescriptionQuery", wraps=DescriptionQuery
            ) as description_query_mock,
        ):
            task = asyncio.create_task(request_description(remote_addr[0]))
            await time_travel(0)
            transport_connect_mock.assert_called_once_with()
            assert transport_send_mock.call_count == 1
            assert isinstance(
                transport_send_mock.call_args[0][0].body, DescriptionRequest
            )

            _transport = description_query_mock.call_args.kwargs["transport"]
            _transport.data_received_callback(
                self.description_response_raw, remote_addr
            )
            await time_travel(0)
            assert transport_send_mock.call_count == 2
            assert isinstance(
                transport_send_mock.call_args[0][0].body, SearchRequestExtended
            )

            _transport = description_query_mock.call_args.kwargs["transport"]
            _transport.data_received_callback(
                self.search_response_extended_raw, remote_addr
            )

            gateway = await task
            transport_stop_mock.assert_called_once()
            assert transport_send_mock.call_count == 2

            assert gateway.core_version == 2
            assert gateway.name == "KNX IP Router 752 secure"
            assert gateway.tunnelling_requires_secure is True
            assert gateway.individual_address == IndividualAddress("5.0.0")



"""Example for internal devices storage."""

import asyncio

from xknx import XKNX
from xknx.devices import Switch


async def main() -> None:
    """Add test Switch to devices storage and access it by name."""
    xknx = XKNX()
    await xknx.start()
    switch = Switch(xknx, name="TestOutlet", group_address="1/1/11")
    xknx.devices.async_add(switch)

    await xknx.devices["TestOutlet"].set_on()
    await asyncio.sleep(2)
    await xknx.devices["TestOutlet"].set_off()
    await xknx.stop()


asyncio.run(main())
